package sugg;

import java.awt.Color;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.Set;

import org.math.plot.Plot2DPanel;
import org.math.plot.plots.Plot;
import org.math.plot.plots.ScatterPlot;
import org.math.plot.render.AbstractDrawer;

import sugg.cluster.ClusterVector;
import sugg.cluster.DocCluster;
import sugg.cluster.QueryCluster;
import sugg.matrix.SMatrix;
import sugg.matrix.SVector;
import sugg.objects.Location;
import sugg.objects.Pair;
import sugg.objects.Query;
import sugg.objects.ResultStat;
import sugg.plot.PlotTools;
import sugg.tools.Comparators;
import sugg.tools.NodeWeighter;
import sugg.tools.SortedMap;
import sugg.tools.SortedMapTreeSet;
import sugg.tools.Timer;
import sugg.tools.Tools;
import sugg.tools.VectorWeighter;

public class BCAWalker {
	enum BCABreakType {BY_THRESHOLD, BY_BOUND, BY_MAX_ITER, BY_THETA, BY_OTHER};
	
	private Timer timer;
	
	private int numQuery;
	private int numDoc;
	
	private SMatrix sQueryDocMatrix;
	private SMatrix sDocQueryMatrix;
	private SMatrix sRawQueryDocMatrix;

	String [] idToStringArray;
	Location [] docToLocArray;
	String [] docToTextArray;
	Location [] cellIdToLocArray;
	Map<String, List<Integer>> queryStringToClusterMap;
	Map<String, Integer> queryStringToIdMap;

	String resultPathPrefix;
	
	boolean AOL;
	
	public BCAWalker (boolean AOL, SMatrix sQueryDocMatrix, SMatrix sDocQueryMatrix, SMatrix sRawQueryDocMatrix,
			int numQuery, int numDoc, String [] idToStringArray, Location [] docToLocArray, 
			String [] docToTextArray, Location [] cellIdToLocArray,
			Map<String, List<Integer>> queryStringToClusterMap, 
			Map<String, Integer> queryStringToIdMap,
			String resultPathPrefix)
	{
		this.AOL = AOL;
		//sQueryDocMatrix, sDocQueryMatrix, sDocDocMatrix
		this.sQueryDocMatrix = sQueryDocMatrix;
		this.sDocQueryMatrix = sDocQueryMatrix;
		this.sRawQueryDocMatrix = sRawQueryDocMatrix;
		this.resultPathPrefix = resultPathPrefix;
		
		this.initialize(numQuery, numDoc, idToStringArray, docToLocArray, docToTextArray, cellIdToLocArray, 
				queryStringToClusterMap, queryStringToIdMap);
	}
	
	private void initialize(int numQuery, int numDoc, 
			String [] idToStringArray, Location [] docToLocArray, 
			String [] docToTextArray, Location [] cellIdToLocArray,
			Map<String, List<Integer>> queryStringToClusterMap,
			Map<String, Integer> queryStringToIdMap)
	{
		this.timer = new Timer();
		this.numQuery = numQuery;
		this.numDoc = numDoc;
		this.idToStringArray = idToStringArray;
		this.docToLocArray = docToLocArray;
		this.docToTextArray = docToTextArray;
		this.cellIdToLocArray = cellIdToLocArray;
		this.queryStringToClusterMap = queryStringToClusterMap;
		this.queryStringToIdMap = queryStringToIdMap;
	}
	
	public ResultStat tranBCA(Query inputQuery, 
			double alpha, double beta, double gamma, double threshold, double range,
			int queryBatch, String resultPath,
			List<Integer> truthList) throws IOException
	{
		int inputId = inputQuery.queryId;
		Location queryLoc = inputQuery.queryLoc;
		
		int k = 10;
		this.timer.start();

		SortedMap<Integer, Double> sortedResult = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		SortedMap<Integer, Double> sortedQuery = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		sortedQuery.put(inputId, 1.0, true);

		NodeWeighter weighter = new NodeWeighter(numQuery, numDoc, queryLoc, docToLocArray, gamma, sQueryDocMatrix);
		
//		double [] docDistArray = this.getDocDistArray(queryLoc);
//		double [] docWeight = this.getDocWeight(docDistArray, gamma);
//		double [] queryDistArray = this.getQueryDistArray(queryLoc, docDistArray, sQueryDocMatrix);
//		double [] queryWeight = this.getQueryWeight(queryDistArray, gamma);

		int iter = 0;
		Timer testTimer = new Timer();
		
		Map<Integer, SVector> queryDocVectorMap = new HashMap<Integer, SVector>();
		Map<Integer, SVector> docQueryVectorMap = new HashMap<Integer, SVector>();
		Map<Integer, SVector> queryQueryVectorMap = new HashMap<Integer, SVector>();
		
		double rInk = 1;
		while(true)
		{
//			Set<Integer> lazySet = new HashSet<Integer>();
//			for (int i = 0; i != sortedQuery.size(); ++i)
//			{
//				int qId = sortedQuery.getKey(i);
//				double val = sortedQuery.getValue(i);
//				if (val <= threshold)
//					lazySet.add(qId);
//			}
//			for (Integer key : lazySet)
//				sortedQuery.put(key, sortedQuery.get(key), false);
//			
//			if (sortedQuery.empty() == true)
//			{
//				System.out.println("Break by Epsilon");
//				break;
//			}
//			sortedQuery.sort();
			
//			if (sortedQuery.topValue() <= threshold)
//				break;
			iter++;
			
			sortedQuery.sort();
			
			//System.out.println("iter" + iter + ": " + sortedQuery.topValue() + " rInk = " + rInk);
			
			testTimer.start();
			
			boolean stop = true;
			for(int ii = 0; ii != queryBatch; ++ii)
			{
				if (sortedQuery.empty())
					break;

				//query
				int qId = sortedQuery.topKey();
				double paint = sortedQuery.topValue();
				
				if (paint <= threshold)
					break;
				
				sortedQuery.pop();
				
				stop = false;
				
				//assert(paint > threshold);

				if (sortedResult.get(qId) != null)
					sortedResult.put(qId, sortedResult.get(qId) + alpha * paint, true);
				else
					sortedResult.put(qId, alpha * paint, true);
				
				rInk -= alpha * paint;

				if (queryQueryVectorMap.containsKey(qId) == false)
				{
					SVector queryQueryVector = new SVector();

					SVector queryDocVector = VectorWeighter.sGetQueryDocVector(qId, 
							sQueryDocMatrix, weighter, beta, range, queryDocVectorMap);

					for (Entry<Integer, Double> docEntry : queryDocVector)
					{
						int docId = docEntry.getKey();
						if (docId >= this.numDoc)
							continue;

						double docInk = docEntry.getValue();
						
						if (docInk < threshold)
							continue;

						SVector docQueryVector = VectorWeighter.sGetDocQueryVector(docId, 
								sDocQueryMatrix, weighter, beta, range, docQueryVectorMap);
						
						for (Entry<Integer, Double> queryEntry : docQueryVector)
						{
							int phraseId = queryEntry.getKey();
							if (phraseId >= this.numQuery)
								continue;
							
							double qQi = queryQueryVector.get(phraseId) + docInk * queryEntry.getValue();
							queryQueryVector.set(phraseId, qQi);
						}
					}
					
					//assert(queryQueryVector.normL1() < 1.0 + Tools.EP);
					queryQueryVectorMap.put(qId, queryQueryVector);
				}
				
				SVector queryQueryVector = queryQueryVectorMap.get(qId);

				paint *= (1 - alpha);
				for (Entry<Integer, Double> queryEntry : queryQueryVector)
				{
					int phraseId = queryEntry.getKey();
					double weight = queryEntry.getValue();
					
					double qQi = sortedQuery.get(phraseId) == null ? 0 : sortedQuery.get(phraseId);
					qQi += paint * weight;
					sortedQuery.put(phraseId, qQi, true);
				}
			}
			
			if (stop)
			{
				System.out.println("Break by Epsilon");
				break;
			}
		}
		
		double cost = this.timer.stop();
		
		ResultStat stat = this.writeResult(alpha, beta, gamma, threshold, range, k,
				inputQuery, weighter, truthList, sortedResult, null, null, 0, 0,
				cost, iter, rInk, resultPath);
		
		return stat;
	}
	
	public void testClusterBCA(int inputId, Location queryLoc, 
			int k, double alpha, double beta, double gamma, double threshold, double range, 
			int [] docClusterArray, int [] queryClusterArray,
			int numDocCluster, int numQueryCluster) throws IOException
	{
		double cost = 0;
		
		this.timer.start();
		
		SortedMap<Integer, QueryCluster> sortedCluster = new SortedMap<Integer, QueryCluster>(
				new Comparator<QueryCluster>() {
					@Override
					public int compare(QueryCluster a, QueryCluster b) {
						return a.compareTo(b);
					}
		});
		for (int i = 0; i != numQueryCluster; ++i)
			sortedCluster.put(i, new QueryCluster(i), false);
		
		SortedMap<Integer, DocCluster> sortedCell = new SortedMap<Integer, DocCluster>(
				new Comparator<DocCluster>() {
					@Override
					public int compare(DocCluster a, DocCluster b) {
						return a.compareTo(b);
					}
		});
		for (int i = 0; i != numDocCluster; ++i)
			sortedCell.put(i, new DocCluster(i), false);
		
		QueryCluster firstCluster = sortedCluster.get(queryClusterArray[inputId]);
		firstCluster.tempVector.set(inputId, 1.0);
		firstCluster.rInk = 1;
		firstCluster.aInk = 1;
		firstCluster.upperBound = 1;
		sortedCluster.put(queryClusterArray[inputId], firstCluster, true);
		
		SortedMap<Integer, Double> sortedResult = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		
		Map<Integer, SVector> queryDocVectorMap = new HashMap<Integer, SVector>();
		Map<Integer, SVector> docQueryVectorMap = new HashMap<Integer, SVector>();
		
		Map<Integer, ClusterVector> queryCellVectorMap = new HashMap<Integer, ClusterVector>();
		Map<Integer, ClusterVector> docClusterVectorMap = new HashMap<Integer, ClusterVector>();
		
		NodeWeighter nodeWeighter = new NodeWeighter(numQuery, numDoc, queryLoc, docToLocArray, gamma, sQueryDocMatrix);
		
		SVector vistedDocVector = new SVector();
		
		double rInk = 1;
		while (true)
		{			
			sortedCluster.sort();

			boolean stop = true;

			if (!sortedCluster.empty())// && sortedCluster.topValue().upperBound > threshold)
			{
				QueryCluster topCluster = sortedCluster.topValue();
				
				//System.out.println("cluster" + topCluster.clusterId);

				boolean alive = topCluster.distribute(inputId, alpha, beta, threshold, range, numQuery, numDoc,
						sortedResult, sortedCell, sQueryDocMatrix, sDocQueryMatrix, 
						queryCellVectorMap, docClusterVectorMap,
						queryDocVectorMap, docQueryVectorMap,
						queryClusterArray, docClusterArray,
						//queryWeight, docWeight, queryDistArray, docDistArray
						nodeWeighter);

				if (alive)
				{
					stop = false;
					for (int i = 0; i < numDocCluster; ++i)
					{
						DocCluster cell = sortedCell.get(i);
						sortedCell.put(i, cell, cell.upperBound > threshold);
					}
				}
				rInk -= topCluster.distInk;
			}

			sortedCell.sort();
			
			if (!sortedCell.empty())// && sortedCell.topValue().upperBound > threshold)
			{
				DocCluster topCell = sortedCell.topValue();
				
				//System.out.println("cell" + topCell.cellId);

				boolean alive = topCell.distribute(alpha, beta, threshold, range, numQuery, numDoc,
						sortedCluster, sQueryDocMatrix, sDocQueryMatrix, 
						queryCellVectorMap, docClusterVectorMap, 
						queryDocVectorMap, docQueryVectorMap,
						queryClusterArray, docClusterArray, 
						//queryWeight, docWeight, queryDistArray, docDistArray
						nodeWeighter,
						vistedDocVector);

				if (alive)
				{
					stop = false;
					for (int i = 0; i < numQueryCluster; ++i)
					{
						QueryCluster cluster = sortedCluster.get(i);
						sortedCluster.put(i, cluster, cluster.upperBound > threshold);
					}
				}
			}
			
			if (stop)
			{
				System.out.println("Break by Epsilon");
				break;
			}
		}
		sortedResult.sort();
		cost += this.timer.stop();
		System.out.println("Cost = " + cost);
		System.out.println("rInk = " + rInk);
		
		List<Integer> connectedDoc = new ArrayList<Integer>();
		for (Entry<Integer, Double> docEntry : sQueryDocMatrix.getRowVector(inputId))
		{
			int docId = docEntry.getKey();
			connectedDoc.add(docId);
		}
		
		double [][] connectedLoc = new double[connectedDoc.size()][2];
		int ii = 0;
		for (Integer docId : connectedDoc)
		{
			Location loc = docToLocArray[docId];
			connectedLoc[ii][0] = loc.getLong();
			connectedLoc[ii][1] = loc.getLat();
			ii++;
		}
		Plot connectedDocPlot = new ScatterPlot("Connected Docs", Color.MAGENTA, AbstractDrawer.ROUND_DOT, 3, connectedLoc);
		

		List<Integer> freqDoc = new ArrayList<Integer>();
		for (Entry<Integer, Double> docEntry : vistedDocVector)
		{
			if (docEntry.getValue() >= 3)
				freqDoc.add(docEntry.getKey());
		}
		
		double [][] visitedLoc = new double[freqDoc.size()][2];
		ii = 0;
		for (Integer docId : freqDoc)
		{
			Location loc = docToLocArray[docId];
			visitedLoc[ii][0] = loc.getLong();
			visitedLoc[ii][1] = loc.getLat();
			ii++;
		}
		Plot visitedDocPlot = new ScatterPlot("Visted Docs", Color.BLUE, AbstractDrawer.ROUND_DOT, 3, visitedLoc);
		
		Set<Integer> resultDocSet = new HashSet<Integer>();
		
		for (int i = 1; i < Math.min(sortedResult.size(), k + 1); ++i)
		{
			int phraseId = sortedResult.getKey(i);
			double score = sortedResult.getValue(i);
			String phrase = idToStringArray[phraseId];

			SVector docVector = sQueryDocMatrix.getRowVector(phraseId);
			for (Entry<Integer, Double> docEntry : docVector)
			{
				int docId = docEntry.getKey();

				double distance = nodeWeighter.getDocDist(docId);
								
				if (distance <= 0.05)
					resultDocSet.add(docId);				
			}
			
			System.out.printf("[%d] %s: %.3f\n", i, phrase, score);
		}
		
		double [][] resultLoc = new double[resultDocSet.size()][2];
		ii = 0;
		for (Integer docId : resultDocSet)
		{
			resultLoc[ii][0] = docToLocArray[docId].getLong();
			resultLoc[ii][1] = docToLocArray[docId].getLat();
			ii++;
		}
		
//		String query = idToStringArray[inputId];
//		double lng = queryLoc.getLong();
//		double lat = queryLoc.getLat();
//		
//		String prefix = query + "_" + lng + "_" + lat + "_b" + beta;
//		
//		String visitedDocPath = plotPathPrefix + prefix + "_visited.txt";
//		PrintWriter writer = new PrintWriter(new File(visitedDocPath));
//		for (int i = 0; i < visitedLoc.length; ++i)
//			writer.println(visitedLoc[i][0] + "\t" + visitedLoc[i][1]);
//		writer.close();
//		
//		String resultDocPath = plotPathPrefix + prefix + "_result.txt";
//		writer = new PrintWriter(new File(resultDocPath));
//		for (int i = 0; i < resultLoc.length; ++i)
//			writer.println(resultLoc[i][0] + "\t" + resultLoc[i][1]);
//		writer.close();

		ScatterPlot resultDocPlot = new ScatterPlot("Result Docs", Color.GREEN, 
				AbstractDrawer.ROUND_DOT, 3, resultLoc);
		
		Plot queryLocPlot = new ScatterPlot("Query Location", Color.ORANGE, 
				AbstractDrawer.PATTERN_DOT, 6, new double[][]{{queryLoc.getLong(), queryLoc.getLat()}});

		Plot2DPanel plotPanel = new Plot2DPanel();
		if (connectedDoc.size() > 0)
			plotPanel.addPlot(connectedDocPlot);
		if (freqDoc.size() > 0)
			plotPanel.addPlot(visitedDocPlot);
		plotPanel.addPlot(queryLocPlot);
		if (resultDocSet.size() > 0)
			plotPanel.addPlot(resultDocPlot);
		plotPanel.setFixedBounds(0, 0, 1);
		plotPanel.setFixedBounds(1, 0, 1);
		
		PlotTools.drawPlot(plotPanel, "Plot");
	}
	
	public void testBCA(int inputId, Location queryLoc, 
			int k, double alpha, double beta, double gamma, double threshold, double range, String plotPathPrefix) throws IOException
	{
		double cost = 0;
		
		this.timer.start();
		
		SVector queryVec = new SVector();
		queryVec.set(inputId, 1.0);
		SVector docVec = new SVector();
		SortedMap<Integer, Double> sortedResult = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());

		NodeWeighter weighter = new NodeWeighter(numQuery, numDoc, queryLoc, docToLocArray, gamma, sQueryDocMatrix);
		
		int iter = 0;
		
		Map<Integer, SVector> queryDocVectorMap = new HashMap<Integer, SVector>();
		Map<Integer, SVector> docQueryVectorMap = new HashMap<Integer, SVector>();

		Set<Integer> querySet = new HashSet<Integer>();
		SVector vistedDocVector = new SVector();
		
		System.out.println("QueryId = " + inputId);
		System.out.println("QueryLoc = " + queryLoc.getLocx() + ", " + queryLoc.getLocy());
		
		double rInk = 1;
		while(true)
		{
			iter++;

			boolean stop = true;

			for (Entry<Integer, Double> queryInkEntry : queryVec)
			{
				int qId = queryInkEntry.getKey();
				double paint = queryInkEntry.getValue();

				if (paint <= threshold)
					continue;
				
				querySet.add(qId);
				
				stop = false;
				
				queryInkEntry.setValue(0.0);


				if (sortedResult.get(qId) != null)
					sortedResult.put(qId, sortedResult.get(qId) + alpha * paint, true);
				else
					sortedResult.put(qId, alpha * paint, true);
				rInk -= alpha * paint;
					
				paint *= (1 - alpha);

				SVector queryDocVector = VectorWeighter.sGetQueryDocVector(qId, 
						sQueryDocMatrix, weighter, beta, range, queryDocVectorMap);

				for (Entry<Integer, Double> docEntry : queryDocVector)
				{
					int docId = docEntry.getKey();
					
					double weight = docEntry.getValue();

					docVec.add(docId, paint * weight);
				}
			}

			for (Entry<Integer, Double> docInkEntry : docVec)
			{
				//doc
				
				int dId = docInkEntry.getKey();
				double ink = docInkEntry.getValue();
				
				if (ink <= threshold)
					continue;
				
				stop = false;
				
				vistedDocVector.add(dId, 1);
				
				docInkEntry.setValue(0.0);

				SVector docQueryVector = VectorWeighter.sGetDocQueryVector(dId, 
						sDocQueryMatrix, weighter, beta, range, docQueryVectorMap);
				
				for (Entry<Integer, Double> queryEntry : docQueryVector)
				{
					int phraseId = queryEntry.getKey();
					
					double weight = queryEntry.getValue();
					
					queryVec.add(phraseId, ink * weight);
				}
			}
			
			if (stop)
			{
				System.out.println("VisitedQuery = " + querySet.size() + " VisitedDoc = " + vistedDocVector.numEntries());
				System.out.println("Break by Epsilon: iter = " + iter);
				break;
			}
			
			if (sortedResult.size() > k)
			{
				double rInkQ = queryVec.normL1();
				sortedResult.sort();
				
				double kth = sortedResult.getValue(k);
				double k1th = sortedResult.getValue(k);
				
				if (k1th + alpha * rInkQ <= kth)
				{
					System.out.println("Break by Bound: iter = " + iter);
					break;
				}
			}
		}
		cost += this.timer.stop();
		System.out.println("Cost = " + cost);
		System.out.println("rInk = " + rInk);
		
		List<Integer> connectedDoc = new ArrayList<Integer>();
		for (Entry<Integer, Double> docEntry : sQueryDocMatrix.getRowVector(inputId))
		{
			int docId = docEntry.getKey();
			connectedDoc.add(docId);
		}
		
		double [][] connectedLoc = new double[connectedDoc.size()][2];
		int ii = 0;
		for (Integer docId : connectedDoc)
		{
			Location loc = docToLocArray[docId];
			connectedLoc[ii][0] = loc.getLong();
			connectedLoc[ii][1] = loc.getLat();
		}
		Plot connectedDocPlot = new ScatterPlot("Connected Docs", Color.MAGENTA, AbstractDrawer.ROUND_DOT, 3, connectedLoc);
		

		List<Integer> freqDoc = new ArrayList<Integer>();
		for (Entry<Integer, Double> docEntry : vistedDocVector)
		{
			if (docEntry.getValue() >= 3)
				freqDoc.add(docEntry.getKey());
		}
		
		double [][] visitedLoc = new double[freqDoc.size()][2];
		ii = 0;
		for (Integer docId : freqDoc)
		{
			Location loc = docToLocArray[docId];
			visitedLoc[ii][0] = loc.getLong();
			visitedLoc[ii][1] = loc.getLat();
			ii++;
		}
		Plot visitedDocPlot = new ScatterPlot("Visted Docs", Color.BLUE, AbstractDrawer.ROUND_DOT, 3, visitedLoc);
//		double [] visitedWeight = new double[freqDoc.size()];
//		for (Entry<Integer, Double> docEntry : vistedDocVector)
//		{
//			int docId = docEntry.getKey();
//			visitedWeight[ii] = docEntry.getValue();
//			
//			Location loc = docToLocArray[docId];
//			
//			visitedLoc[ii][0] = loc.getLong();
//			visitedLoc[ii][1] = loc.getLat();
//			ii++;
//		}
//		Tools.normalizeArrayTo01(visitedWeight);
//		Plot visitedDocPlot = new MyScatterPlot("Visited Docs", Color.BLUE, AbstractDrawer.ROUND_DOT, 3, visitedLoc, visitedWeight);
		
		Set<Integer> resultDocSet = new HashSet<Integer>();
		
		for (int i = 1; i < Math.min(sortedResult.size(), k + 1); ++i)
		{
			int phraseId = sortedResult.getKey(i);
			double score = sortedResult.getValue(i);
			String phrase = idToStringArray[phraseId];

			SVector docVector = sQueryDocMatrix.getRowVector(phraseId);
			for (Entry<Integer, Double> docEntry : docVector)
			{
				int docId = docEntry.getKey();

				double distance = weighter.getDocDist(docId);
								
				if (distance <= 0.05)
					resultDocSet.add(docId);				
			}
			
			System.out.printf("[%d] %s: %.3f\n", i, phrase, score);
		}
		
		double [][] resultLoc = new double[resultDocSet.size()][2];
		ii = 0;
		for (Integer docId : resultDocSet)
		{
			resultLoc[ii][0] = docToLocArray[docId].getLong();
			resultLoc[ii][1] = docToLocArray[docId].getLat();
			ii++;
		}
		
		String query = idToStringArray[inputId];
		double lng = queryLoc.getLong();
		double lat = queryLoc.getLat();		
		String prefix = query + "_" + lng + "_" + lat + "_b" + beta;

		String visitedDocPath = plotPathPrefix + prefix + "_visited.txt";
		PrintWriter writer = new PrintWriter(new File(visitedDocPath));
		for (int i = 0; i < visitedLoc.length; ++i)
			writer.println(visitedLoc[i][0] + "\t" + visitedLoc[i][1]);
		writer.close();
		
		String resultDocPath = plotPathPrefix + prefix + "_result.txt";
		writer = new PrintWriter(new File(resultDocPath));
		for (int i = 0; i < resultLoc.length; ++i)
			writer.println(resultLoc[i][0] + "\t" + resultLoc[i][1]);
		writer.close();

		ScatterPlot resultDocPlot = new ScatterPlot("Result Docs", Color.GREEN, 
				AbstractDrawer.ROUND_DOT, 3, resultLoc);
		
		Plot queryLocPlot = new ScatterPlot("Query Location", Color.ORANGE, 
				AbstractDrawer.PATTERN_DOT, 6, new double[][]{{queryLoc.getLong(), queryLoc.getLat()}});

		Plot2DPanel plotPanel = new Plot2DPanel();
		if (connectedDoc.size() > 0)
			plotPanel.addPlot(connectedDocPlot);
		if (freqDoc.size() > 0)
			plotPanel.addPlot(visitedDocPlot);
		plotPanel.addPlot(queryLocPlot);
		if (resultDocSet.size() > 0)
			plotPanel.addPlot(resultDocPlot);
		plotPanel.setFixedBounds(0, 0, 1);
		plotPanel.setFixedBounds(1, 0, 1);
		
		PlotTools.drawPlot(plotPanel, "Plot");
	}
	
	public ResultStat batchBCA(Query inputQuery, 
			int k, double alpha, double beta, double gamma, double threshold, double range,
			int queryBatch, int docBatch,
			String resultPath,
			List<Integer> truthList) throws IOException
	{
		int inputId = inputQuery.queryId;
		Location queryLoc = inputQuery.queryLoc;
		
		this.timer.start();
		
		SortedMap<Integer, Double> sortedResult = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		SortedMap<Integer, Double> sortedQuery = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		sortedQuery.put(inputId, 1.0, true);
		SortedMap<Integer, Double> sortedDoc = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		
		NodeWeighter weighter = new NodeWeighter(numQuery, numDoc, queryLoc, docToLocArray, gamma, sQueryDocMatrix);
		
		int iter = 0;
		
		Map<Integer, SVector> queryDocVectorMap = new HashMap<Integer, SVector>();
		Map<Integer, SVector> docQueryVectorMap = new HashMap<Integer, SVector>();

		double rInk = 1;
		while(true)
		{
			if (sortedQuery.empty() == true && sortedDoc.empty() == true)
				break;

			iter++;

			boolean stop = true;

			sortedQuery.sort();
			
			for(int ii = 0; ii != queryBatch; ++ii)
			{
				if (sortedQuery.empty())
					break;

				//query
				int qId = sortedQuery.topKey();
				double paint = sortedQuery.topValue();

				if (paint <= threshold)
					break;
				
				stop = false;
				
				sortedQuery.pop();
								
				//assert(paint > threshold);

				if (sortedResult.get(qId) != null)
					sortedResult.put(qId, sortedResult.get(qId) + alpha * paint, true);
				else
					sortedResult.put(qId, alpha * paint, true);
				
				rInk -= alpha * paint;
				
				//System.out.print("Query to Doc: ");
				SVector queryDocVector = VectorWeighter.sGetQueryDocVector(qId, 
						sQueryDocMatrix, weighter, beta, range, queryDocVectorMap);
//				SVector queryDocVector = WeightedVectorTools.sGetQueryDocVector(qId, 
//						sQueryDocMatrix, this.numDoc, docWeight, beta, docDistArray, range,
//						queryDocVectorMap);

				for (Entry<Integer, Double> docEntry : queryDocVector)
				{
					int docId = docEntry.getKey();
					if (docId >= this.numDoc)
						continue;
					
					double weight = docEntry.getValue();

					double qDj = sortedDoc.get(docId) == null ? 0 : sortedDoc.get(docId);
					qDj += (1 - alpha) * paint * weight;
					sortedDoc.put(docId, qDj, qDj > threshold);
				}
			}
			
			sortedDoc.sort();
			
			for(int ii = 0; ii != docBatch; ++ii)
			{
				//doc
				if (sortedDoc.empty())
					break;
				
				int dId = sortedDoc.topKey();
				double ink = sortedDoc.topValue();
				
				if (ink <= threshold)
					break;
				
				stop = false;
				
				sortedDoc.pop();

				//System.out.print("Doc to Query: ");

				SVector docQueryVector = VectorWeighter.sGetDocQueryVector(dId, 
						sDocQueryMatrix, weighter, beta, range, docQueryVectorMap);
				
				for (Entry<Integer, Double> queryEntry : docQueryVector)
				{
					int phraseId = queryEntry.getKey();
					if (phraseId >= this.numQuery)
						continue;
					
					double weight = queryEntry.getValue();
					
					double qQi = sortedQuery.get(phraseId) == null ? 0 : sortedQuery.get(phraseId);
					qQi +=  ink * weight;
					sortedQuery.put(phraseId, qQi, qQi > threshold);
				}
			}
			
			if (stop)
			{
				System.out.println("Break by Epsilon");
				break;
			}
		}
		double cost = this.timer.stop();
		
		if (sortedResult.size() <= k)
			return null;
		
		ResultStat stat = this.writeResult(alpha, beta, gamma, threshold, range, k,
				inputQuery, weighter, truthList, sortedResult, null, null, 0, 0,
				cost, iter, rInk, resultPath);
		
		return stat;
	}
	
	public ResultStat batchBCA_MoreEdge(Query inputQuery, 
			int k, double alpha, double beta, double gamma, double threshold, double range,
			int queryBatch, int docBatch,
			String resultPath,
			SMatrix sDocDocMatrix) throws IOException
	{
		int inputId = inputQuery.queryId;
		Location queryLoc = inputQuery.queryLoc;
		
		this.timer.start();
		
		SortedMap<Integer, Double> sortedResult = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		SortedMap<Integer, Double> sortedQuery = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		sortedQuery.put(inputId, 1.0, true);
		SortedMap<Integer, Double> sortedDoc = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		
		NodeWeighter weighter = new NodeWeighter(numQuery, numDoc, queryLoc, docToLocArray, gamma, sQueryDocMatrix);
		
		int iter = 0;
		
		Map<Integer, SVector> queryDocVectorMap = new HashMap<Integer, SVector>();
		Map<Integer, SVector> docQueryVectorMap = new HashMap<Integer, SVector>();

		double ratio = 0.5;
		double rInk = 1;
		while(true)
		{
			if (sortedQuery.empty() == true && sortedDoc.empty() == true)
				break;

			iter++;

			boolean stop = true;

			sortedQuery.sort();
			
			for(int ii = 0; ii != queryBatch; ++ii)
			{
				if (sortedQuery.empty())
					break;

				//query
				int qId = sortedQuery.topKey();
				double paint = sortedQuery.topValue();

				if (paint <= threshold)
					break;
				
				stop = false;
				
				sortedQuery.pop();
								
				//assert(paint > threshold);

				if (sortedResult.get(qId) != null)
					sortedResult.put(qId, sortedResult.get(qId) + alpha * paint, true);
				else
					sortedResult.put(qId, alpha * paint, true);
				
				rInk -= alpha * paint;
				
				//System.out.print("Query to Doc: ");
				SVector queryDocVector = VectorWeighter.sGetQueryDocVector(qId, 
						sQueryDocMatrix, weighter, beta, range, queryDocVectorMap);
//				SVector queryDocVector = WeightedVectorTools.sGetQueryDocVector(qId, 
//						sQueryDocMatrix, this.numDoc, docWeight, beta, docDistArray, range,
//						queryDocVectorMap);

				for (Entry<Integer, Double> docEntry : queryDocVector)
				{
					int docId = docEntry.getKey();
					if (docId >= this.numDoc)
						continue;
					
					double weight = docEntry.getValue();

					double qDj = sortedDoc.get(docId) == null ? 0 : sortedDoc.get(docId);
					qDj += (1 - alpha) * paint * weight;
					sortedDoc.put(docId, qDj, qDj > threshold);
				}
			}
			
			sortedDoc.sort();
			
			for(int ii = 0; ii != docBatch; ++ii)
			{
				//doc
				if (sortedDoc.empty())
					break;
				
				int dId = sortedDoc.topKey();
				double ink = sortedDoc.topValue();
				
				if (ink <= threshold)
					break;
				
				stop = false;
				
				sortedDoc.pop();

				//System.out.print("Doc to Query: ");

				SVector docQueryVector = VectorWeighter.sGetDocQueryVector(dId, 
						sDocQueryMatrix, weighter, beta, range, docQueryVectorMap);
//				SVector docQueryVector = WeightedVectorTools.sGetDocQueryVector(dId, 
//						sDocQueryMatrix, this.numQuery, queryWeight, beta, queryDistArray, range, docQueryVectorMap);

				SVector docDocVector = sDocDocMatrix.getRowVector(dId);
				
				for (Entry<Integer, Double> queryEntry : docQueryVector)
				{
					int phraseId = queryEntry.getKey();
					if (phraseId >= this.numQuery)
						continue;
					
					double weight = queryEntry.getValue();
					
					double qQi = sortedQuery.get(phraseId) == null ? 0 : sortedQuery.get(phraseId);
					if (docDocVector == null || docDocVector.size() == 0)
						qQi += ink * weight;
					else
						qQi +=  ratio * ink * weight;
					sortedQuery.put(phraseId, qQi, qQi > threshold);
				}
				
				for (Entry<Integer, Double> docEntry : docDocVector)
				{
					int docId = docEntry.getKey();
					if (docId >= this.numDoc)
						continue;
					
					double weight = docEntry.getValue();
					double qDj = sortedDoc.get(docId) == null ? 0 : sortedDoc.get(docId);
					qDj += (1.0 - ratio) * ink * weight;
					sortedDoc.put(docId, qDj, qDj > threshold);
				}
			}
			
			if (stop)
			{
				System.out.println("Break by Epsilon");
				break;
			}
		}
		double cost = this.timer.stop();
		
		if (sortedResult.size() <= k)
			return null;
		
		ResultStat stat = this.writeResult(alpha, beta, gamma, threshold, range, k,
				inputQuery, weighter, null, sortedResult, null, null, 0, 0,
				cost, iter, rInk, resultPath);
		
		return stat;
	}
	
	public ResultStat clusterBCA_Old(Query query, 
			int k, double alpha, double beta, double gamma, double threshold, double range, 
			int [] docClusterArray, int [] queryClusterArray,
			int numDocCluster, int numQueryCluster,
			String resultPath,
			List<Integer> truthList) throws IOException
	{
		int inputId = query.queryId;
		Location queryLoc = query.queryLoc;
		
		double cost = 0;
		
		this.timer.start();
		
		SortedMap<Integer, QueryCluster> sortedCluster = new SortedMap<Integer, QueryCluster>(
				new Comparator<QueryCluster>() {
					@Override
					public int compare(QueryCluster a, QueryCluster b) {
						return a.compareTo(b);
					}
		});
		for (int i = 0; i != numQueryCluster; ++i)
			sortedCluster.put(i, new QueryCluster(i), false);
		
		SortedMap<Integer, DocCluster> sortedCell = new SortedMap<Integer, DocCluster>(
				new Comparator<DocCluster>() {
					@Override
					public int compare(DocCluster a, DocCluster b) {
						return a.compareTo(b);
					}
		});
		for (int i = 0; i != numDocCluster; ++i)
			sortedCell.put(i, new DocCluster(i), false);
		
		QueryCluster firstCluster = sortedCluster.get(queryClusterArray[inputId]);
		firstCluster.tempVector.set(inputId, 1.0);
		firstCluster.rInk = 1;
		firstCluster.aInk = 1;
		firstCluster.upperBound = 1;
		sortedCluster.put(queryClusterArray[inputId], firstCluster, true);

		SortedMap<Integer, Double> sortedResult = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		
		Map<Integer, SVector> queryDocVectorMap = new HashMap<Integer, SVector>();
		Map<Integer, SVector> docQueryVectorMap = new HashMap<Integer, SVector>();
		
		Map<Integer, ClusterVector> queryCellVectorMap = new HashMap<Integer, ClusterVector>();
		Map<Integer, ClusterVector> docClusterVectorMap = new HashMap<Integer, ClusterVector>();
		
		NodeWeighter nodeWeighter = new NodeWeighter(numQuery, numDoc, queryLoc, docToLocArray, gamma, sQueryDocMatrix);
		
		int iter = 0;
		double rInk = 1;
		while (true)
		{
			iter++;

			sortedCluster.sort();

			boolean stop = true;

			if (!sortedCluster.empty())// && sortedCluster.topValue().upperBound > threshold)
			{
				QueryCluster topCluster = sortedCluster.topValue();
				
				//System.out.println("cluster" + topCluster.clusterId);

				boolean alive = topCluster.distribute(inputId, alpha, beta, threshold, range, numQuery, numDoc,
						sortedResult, sortedCell, sQueryDocMatrix, sDocQueryMatrix, 
						queryCellVectorMap, docClusterVectorMap,
						queryDocVectorMap, docQueryVectorMap,
						queryClusterArray, docClusterArray,
						//queryWeight, docWeight, queryDistArray, docDistArray
						nodeWeighter);

				if (alive)
				{
					stop = false;
					for (int i = 0; i < numDocCluster; ++i)
					{
						DocCluster cell = sortedCell.get(i);
						sortedCell.put(i, cell, cell.upperBound > threshold);
					}
				}
				rInk -= topCluster.distInk;
			}

			sortedCell.sort();
			
			if (!sortedCell.empty())// && sortedCell.topValue().upperBound > threshold)
			{
				DocCluster topCell = sortedCell.topValue();
				
				//System.out.println("cell" + topCell.cellId);

				boolean alive = topCell.distribute(alpha, beta, threshold, range, numQuery, numDoc,
						sortedCluster, sQueryDocMatrix, sDocQueryMatrix, 
						queryCellVectorMap, docClusterVectorMap, 
						queryDocVectorMap, docQueryVectorMap,
						queryClusterArray, docClusterArray, 
						//queryWeight, docWeight, queryDistArray, docDistArray
						nodeWeighter, null);

				if (alive)
				{
					stop = false;
					for (int i = 0; i < numQueryCluster; ++i)
					{
						QueryCluster cluster = sortedCluster.get(i);
						sortedCluster.put(i, cluster, cluster.upperBound > threshold);
					}
				}
			}
			
			if (stop)
			{
				System.out.println("Break by Epsilon");
				break;
			}
		}
		cost += this.timer.stop();
		
		// Trivial
		//System.out.println("SortCost = " + (sortedCluster.getSortCost() + sortedCell.getSortCost()));
		
//		double [] docDistArray = this.getDocDistArray(queryLoc);
//		double [] docWeight = this.getDocWeight(docDistArray, gamma);
//
//		double [] queryDistArray = this.getQueryDistArray(queryLoc, docDistArray, sQueryDocMatrix);
//		double [] queryWeight = this.getQueryWeight(queryDistArray, gamma);
		
		
		ResultStat stat = this.writeResult(alpha, beta, gamma, threshold, range, k, 
				query, nodeWeighter, truthList, sortedResult, 
				sortedCluster, sortedCell, numQueryCluster, numDocCluster,
				cost, iter, rInk, resultPath);
		
		return stat;
	}
	
	public ResultStat clusterBCA(Query query, 
			int k, double alpha, double beta, double gamma, double threshold, double range, 
			int [] docClusterArray, int [] queryClusterArray,
			int numDocCluster, int numQueryCluster,
			String resultPath,
			List<Integer> truthList) throws IOException
	{
		int inputId = query.queryId;
		Location queryLoc = query.queryLoc;
		
		double cost = 0;
		
		this.timer.start();
		
		SortedMapTreeSet<Integer, QueryCluster> sortedCluster = new SortedMapTreeSet<Integer, QueryCluster>(
				new Comparator<QueryCluster>() {
					@Override
					public int compare(QueryCluster a, QueryCluster b) {
						return a.compareTo(b);
					}
		});
		for (int i = 0; i != numQueryCluster; ++i)
			sortedCluster.put(i, new QueryCluster(i));
			//sortedCluster.put(i, new QueryCluster(i), false);
		
		SortedMapTreeSet<Integer, DocCluster> sortedCell = new SortedMapTreeSet<Integer, DocCluster>(
				new Comparator<DocCluster>() {
					@Override
					public int compare(DocCluster a, DocCluster b) {
						return a.compareTo(b);
					}
		});
		for (int i = 0; i != numDocCluster; ++i)
			sortedCell.put(i, new DocCluster(i));
			//sortedCell.put(i, new DocCluster(i), false);
		
		QueryCluster firstCluster = sortedCluster.get(queryClusterArray[inputId]);
		firstCluster.tempVector.set(inputId, 1.0);
		firstCluster.rInk = 1;
		firstCluster.aInk = 1;
		firstCluster.upperBound = 1;
		sortedCluster.put(queryClusterArray[inputId], firstCluster);
		//sortedCluster.put(queryClusterArray[inputId], firstCluster, true);

		SortedMapTreeSet<Integer, Double> sortedResult = new SortedMapTreeSet<Integer, Double>(new Comparators.CompareDoubleDesc());
		//SortedMap<Integer, Double> sortedResult = new SortedMap<Integer, Double>(new Comparators.CompareDoubleDesc());
		
		Map<Integer, SVector> queryDocVectorMap = new HashMap<Integer, SVector>();
		Map<Integer, SVector> docQueryVectorMap = new HashMap<Integer, SVector>();
		
		Map<Integer, ClusterVector> queryCellVectorMap = new HashMap<Integer, ClusterVector>();
		Map<Integer, ClusterVector> docClusterVectorMap = new HashMap<Integer, ClusterVector>();
		
		NodeWeighter nodeWeighter = new NodeWeighter(numQuery, numDoc, queryLoc, docToLocArray, gamma, sQueryDocMatrix);
		
		int iter = 0;
		double rInk = 1;
		while (true)
		{
			iter++;

			//sortedCluster.sort();

			boolean stop = true;

			if (!sortedCluster.empty())// && sortedCluster.topValue().upperBound > threshold)
			{
				QueryCluster topCluster = sortedCluster.poll().value;
				//QueryCluster topCluster = sortedCluster.topValue();
				
				//System.out.println("cluster" + topCluster.clusterId);

				boolean alive = topCluster.distribute(inputId, alpha, beta, threshold, range, numQuery, numDoc,
						sortedResult, sortedCell, sQueryDocMatrix, sDocQueryMatrix, 
						queryCellVectorMap, docClusterVectorMap,
						queryDocVectorMap, docQueryVectorMap,
						queryClusterArray, docClusterArray,
						nodeWeighter);
				
				sortedCluster.put(topCluster.clusterId, topCluster);

				if (alive)
				{
					stop = false;
					for (int i = 0; i < numDocCluster; ++i)
					{
						DocCluster cell = sortedCell.get(i);
						sortedCell.put(i, cell);
						//sortedCell.put(i, cell, cell.upperBound > threshold);
					}
				}
				rInk -= topCluster.distInk;
			}

			//sortedCell.sort();
			
			if (!sortedCell.empty())// && sortedCell.topValue().upperBound > threshold)
			{
				DocCluster topCell = sortedCell.poll().value;
				//DocCluster topCell = sortedCell.topValue();
				
				//System.out.println("cell" + topCell.cellId);

				boolean alive = topCell.distribute(alpha, beta, threshold, range, numQuery, numDoc,
						sortedCluster, sQueryDocMatrix, sDocQueryMatrix, 
						queryCellVectorMap, docClusterVectorMap, 
						queryDocVectorMap, docQueryVectorMap,
						queryClusterArray, docClusterArray, 
						//queryWeight, docWeight, queryDistArray, docDistArray
						nodeWeighter, null);
				
				sortedCell.put(topCell.cellId, topCell);

				if (alive)
				{
					stop = false;
					for (int i = 0; i < numQueryCluster; ++i)
					{
						QueryCluster cluster = sortedCluster.get(i);
						sortedCluster.put(i, cluster);
						//sortedCluster.put(i, cluster, cluster.upperBound > threshold);
					}
				}
			}
			
			if (stop)
			{
				System.out.println("Break by Epsilon");
				break;
			}
		}
		cost += this.timer.stop();
		
		// Trivial
		//System.out.println("SortCost = " + (sortedCluster.getSortCost() + sortedCell.getSortCost()));
		
//		double [] docDistArray = this.getDocDistArray(queryLoc);
//		double [] docWeight = this.getDocWeight(docDistArray, gamma);
//
//		double [] queryDistArray = this.getQueryDistArray(queryLoc, docDistArray, sQueryDocMatrix);
//		double [] queryWeight = this.getQueryWeight(queryDistArray, gamma);
		
		
		ResultStat stat = this.writeResult(alpha, beta, gamma, threshold, range, k, 
				query, nodeWeighter, truthList, sortedResult, 
				sortedCluster, sortedCell, numQueryCluster, numDocCluster,
				cost, iter, rInk, resultPath);
		
		return stat;
	}
	
	private ResultStat writeResult (
			double alpha, double beta, double gamma, double threshold, double range, int k,
			Query query, NodeWeighter weighter, List<Integer> truthList,
			SortedMap<Integer, Double> sortedResult, 
			SortedMap<Integer, QueryCluster> sortedCluster,
			SortedMap<Integer, DocCluster> sortedCell,
			int numCluster, int numCell,
			double cost, int iter, double rInk,
			String resultPath) throws IOException
	{
		k = 1;
		
		final int RESULT_SIZE = 10;
		final double RANGE = 0.1;
		int inputId = query.queryId;
		String inputString = this.idToStringArray[inputId];
		String [] inputWords = inputString.split(" ");
		
		Location queryLoc = query.queryLoc;
		
		Queue<Pair<Integer, Double>> docQueue = new PriorityQueue<Pair<Integer, Double>>(RESULT_SIZE, 
				new Comparator<Pair<Integer, Double>>() {
					@Override
					public int compare(Pair<Integer, Double> a, Pair<Integer, Double> b) {
						if (a.right > b.right)
							return 1;
						else if (a.right < b.right)
							return -1;
						else
							return 0;
					}
		});
		
		double [] validRange = {0.02, 0.04, 0.06, 0.08, 0.1};
		int [] numValid = new int[validRange.length];
		
		//retrieve top-$RESULT_SIZE documents w.r.t. the original input
		for (Entry<Integer, Double> entry : this.sQueryDocMatrix.getRowVector(inputId))
		{
			int docId = entry.getKey();
			
			double text = 0;
			if (AOL)
				text = this.sRawQueryDocMatrix.get(inputId, docId);
			else
			{
				for (String word : inputWords)
				{
					if (this.queryStringToIdMap.containsKey(word))
					{
						int id = this.queryStringToIdMap.get(word);
						text += this.sRawQueryDocMatrix.get(id, docId);					
					}
				}				
			}
			
			docQueue.add(new Pair<Integer, Double>(docId, text));
			if (docQueue.size() > RESULT_SIZE)
				docQueue.poll();
		}

		List<Double> originalScore = new ArrayList<Double>();
		List<Integer> originalResult = new ArrayList<Integer>();
		Set<String> originalWord = new HashSet<String>();
		while (docQueue.isEmpty() == false)
		{
			Pair<Integer, Double> doc = docQueue.poll();
			int docId = doc.left;
			double score = doc.right;
				
			originalScore.add(score);

			originalResult.add(docId);
			
			if (AOL == false)
			{
				String text = this.docToTextArray[docId];
				for (String word : text.split(" "))
					originalWord.add(word);
			}
		}
		
		sortedResult.sort();
		List<Integer> resultList = new ArrayList<Integer>();
        double tfidfScoreStat = 0;
        double cosScoreStat = 0;
		PrintWriter writer = new PrintWriter(new FileWriter(resultPath, true));
		writer.println(inputId + ": " + this.idToStringArray[inputId] + " (" + queryLoc.getLocx() + ", " + queryLoc.getLocy() + ")");
		writer.println("alpha = " + alpha + " beta = " + beta + " gamma = " + gamma + " epsilon = " + threshold);

		PrintWriter phraseWriter = new PrintWriter(new FileWriter(this.resultPathPrefix + "phrase" + beta + ".txt", true));
		phraseWriter.println(this.idToStringArray[inputId]);
		phraseWriter.println();

		for (int i = 1; i < Math.min(sortedResult.size(), 6); ++i)
		{
			int phraseId = sortedResult.getKey(i);
			String phrase = idToStringArray[phraseId];
			phraseWriter.println(phrase);
		}
		phraseWriter.println();
		phraseWriter.close();
		
        double normSuggests = 1;
        for (int i = 2; i <= k; ++i)
        	normSuggests += 1.0 / Math.log(i);
		
		double expectedQuery = 0;
		double expectedDoc = 0;

		for (int i = 1; i < Math.min(sortedResult.size(), k + 1); ++i)
		{
			int phraseId = sortedResult.getKey(i);
			if (phraseId == query.expectedQueryId)
				expectedQuery += 1;
			
			double score = sortedResult.getValue(i);
			String phrase = idToStringArray[phraseId];
			String [] suggestWords = phrase.split(" ");
			resultList.add(phraseId);

			SVector docVector = sQueryDocMatrix.getRowVector(phraseId);
			//for each document relevant to suggested query
			for (Entry<Integer, Double> docEntry : docVector)
			{
				int docId = docEntry.getKey();
				
				double dist = weighter.getDocDist(docId);
				
				//check the distance of the document
				for (int jj = validRange.length - 1; jj >= 0; --jj)
				{
					//largest to smallest
					if (dist <= validRange[jj])
						numValid[jj]++;
					else
						break;
				}
				
				//score of this document w.r.t. the suggested keyword
				double text = 0;
				if (AOL)
					text = this.sRawQueryDocMatrix.get(phraseId, docId);
				else
				{
					for (String word : suggestWords)
					{
						if (this.queryStringToIdMap.containsKey(word))
						{
							int id = this.queryStringToIdMap.get(word);
							text += this.sRawQueryDocMatrix.get(id, docId);						
						}
					}					
				}
			
				//top documents w.r.t. the suggested keyword
				docQueue.add(new Pair<Integer, Double>(docId, text));
				if (docQueue.size() > RESULT_SIZE)
					docQueue.poll();
			}

//			SVector orgVec = new SVector();
//			SVector newVec = new SVector();
			
			Set<Integer> newResult = new HashSet<Integer>();
			int rank = docQueue.size();
			double cosScore = 0;
			//for each top document
			while (docQueue.isEmpty() == false)
			{
				Pair<Integer, Double> doc = docQueue.poll();
				int docId = doc.left;
								
				if (docId == query.expectedDocId)
					expectedDoc += 1;
				
				if (!AOL && rank <= originalResult.size())
				{
					Set<String> wordSet = new HashSet<String>();
					newResult.add(docId);
					String text = this.docToTextArray[docId];
					for (String word : text.split(" "))
						wordSet.add(word);
					wordSet.addAll(originalWord);
					
					Map<String, Integer> wordIdMap = new HashMap<String, Integer>();
					int ii = 0;
					for (String word : wordSet)
						wordIdMap.put(word, ii++);
					
					SVector orgVec = new SVector();
					SVector newVec = new SVector();
					
					String originalString = this.docToTextArray[originalResult.get(rank - 1)];					
					for (String word : originalString.split(" "))
						orgVec.add(wordIdMap.get(word), 1);

					String newString = this.docToTextArray[docId];
					for (String word : newString.split(" "))
						newVec.add(wordIdMap.get(word), 1);
					
					double cosSim = orgVec.cosineSimilarity(newVec);
					if (rank > 1)
						cosSim /= Math.log(rank);
					cosScore += cosSim;
				}
				
				rank--;
			}
			
			double normDoc = 1;
			for (int ii = 2; ii < docQueue.size(); ++ii)
				normDoc += 1.0 / Math.log(ii);

			//cosScore = orgVec.cosineSimilarity(newVec);
			cosScore /= normDoc;
			
			for (Entry<Integer, Double> docEntry : docVector)
			{
				int docId = docEntry.getKey();
				
				double dist = weighter.getDocDist(docId);
				
				if (dist > RANGE)
					continue;
				
				double text = 0;
				if (AOL)
					text = this.sRawQueryDocMatrix.get(phraseId, docId);
				else
				{
					for (String word : suggestWords)
					{
						if (this.queryStringToIdMap.containsKey(word))
						{
							int id = this.queryStringToIdMap.get(word);
							text += this.sRawQueryDocMatrix.get(id, docId);						
						}
					}					
				}
			
				docQueue.add(new Pair<Integer, Double>(docId, text));
				if (docQueue.size() > RESULT_SIZE)
					docQueue.poll();
			}
			
			normDoc = 1;
			for (int ii = 2; ii < docQueue.size(); ++ii)
				normDoc += 1.0 / Math.log(ii);

			double tfidfScore = 0;
			rank = docQueue.size();
			while (docQueue.isEmpty() == false)
			{
				Pair<Integer, Double> doc = docQueue.poll();
				
				//double tfidfSuggest = this.sRawQueryDocMatrix.get(phraseId, doc.left);//doc.right;
				
				double tfidfSuggest = 0;
				for (String word : inputWords)
				{
					if (this.queryStringToIdMap.containsKey(word))
					{
						int id = this.queryStringToIdMap.get(word);
						tfidfSuggest += this.sRawQueryDocMatrix.get(id, doc.left);
					}
				}
				
				double ratio = tfidfSuggest;
				if (rank <= originalScore.size())
					ratio /= originalScore.get(rank - 1);
				
				if (rank > 1)
					ratio /= Math.log(rank);
				tfidfScore += ratio;
				
				rank--;
			}
			
			tfidfScore /= normDoc;
			
			if (i > 1)
			{
				tfidfScore /= Math.log(i);
				cosScore /= Math.log(i);
			}
			
			tfidfScoreStat += tfidfScore;
			cosScoreStat += cosScore;
			
			writer.println("[" + i + "] " + phrase + "\t" + score + "\t" + tfidfScore + "\t" + cosScore);
			
//			if (this.docToTextArray != null)
//			{
//				for (int ii = docList.size() - 1; ii >= 0; --ii)
//				{
//					int dId = docList.get(ii);
//					writer.println("\t" + this.docToTextArray[dId]);
//					writer.println("\t" + this.docToLocArray[dId].getLocx() + "\t" + this.docToLocArray[dId].getLocy());
//				}
//			}
		}
		
		tfidfScoreStat /= normSuggests;
		cosScoreStat /= normSuggests;
		
//		if (sortedCluster != null)
//		{
//			double distCost = 0, accuCost = 0;
//			for (int i = 0; i != numCluster; ++i)
//			{
//				QueryCluster cluster = sortedCluster.get(i);
//				writer.printf("%d(%.4f) ", cluster.statNumDist, cluster.statInk);
//				distCost += cluster.distCost;
//				accuCost += cluster.accuCost;
//			}
//			writer.println();
//			writer.println("DistCost = " + distCost);
//			writer.println("AccuCost = " + accuCost);
//		}
//		
//		if (sortedCell != null)
//		{
//			double distCost = 0, accuCost = 0;
//			for (int i = 0; i != numCluster; ++i)
//			{
//				DocCluster cell = sortedCell.get(i);
//				writer.printf("%d(%.4f) ", cell.statNumDist, cell.statInk);
//				distCost += cell.distCost;
//				accuCost += cell.accuCost;
//			}
//			writer.println();
//			writer.println("DistCost = " + distCost);
//			writer.println("AccuCost = " + accuCost);
//		}
		
		writer.println("textScore = " + tfidfScoreStat + "\tdocSim = " + cosScoreStat);
		for (int jj = 0; jj < validRange.length; ++jj)
			writer.println("numValid(" + validRange[jj] + ") = " + numValid[jj]);
		writer.println("rInk = " + rInk);
		writer.println("iter = " + iter);
		writer.println("cost = " + cost);
		writer.println("****************************");
		writer.println();
		writer.close();

		double similarityToTruth = 0;
		double misMatch = 0;
		if (truthList != null)
		{
			if (truthList.isEmpty())
			{
				truthList.addAll(resultList);
				similarityToTruth = 1;
				misMatch = 0;
			}
			else
			{
				misMatch = Tools.countMismatch(resultList, truthList);
				//textScoreStat = Tools.jaccardSimilarity(resultList, truthList);
				//textScoreStat = Tools.getNDCG(resultList, truthList);
				similarityToTruth = Tools.averagePrecision(resultList, truthList);
			}			
		}

		ResultStat stat = new ResultStat();
		stat.add(cost);
		stat.add(iter);
		stat.add(tfidfScoreStat);
		stat.add(similarityToTruth);
		stat.add(misMatch);
		stat.add(cosScoreStat);
		stat.add(rInk);
		for (double num : numValid)
			stat.add(num / k);
		stat.add(expectedQuery);
		stat.add(expectedDoc);
		
		return stat;
	}
	
	private ResultStat writeResult (
			double alpha, double beta, double gamma, double threshold, double range, int k,
			Query query, NodeWeighter weighter, List<Integer> truthList,
			SortedMapTreeSet<Integer, Double> sortedResult, 
			SortedMapTreeSet<Integer, QueryCluster> sortedCluster,
			SortedMapTreeSet<Integer, DocCluster> sortedCell,
			int numCluster, int numCell,
			double cost, int iter, double rInk,
			String resultPath) throws IOException
	{
		k = 1;
		
		final int RESULT_SIZE = 5;
		final double RANGE = 0.1;
		int inputId = query.queryId;
		String inputString = this.idToStringArray[inputId];
		String [] inputWords = inputString.split(" ");
		
		Location queryLoc = query.queryLoc;
		
		class PairComparator implements Comparator<Pair<Integer, Double>> {
			@Override
			public int compare(Pair<Integer, Double> a, Pair<Integer, Double> b) {
				if (a.right > b.right)
					return 1;
				else if (a.right < b.right)
					return -1;
				else
					return 0;
			}
		}
		
		Queue<Pair<Integer, Double>> docQueue = new PriorityQueue<Pair<Integer, Double>>(RESULT_SIZE, new PairComparator());
		
		Queue<Pair<Integer, Double>> nearbyTopDocQueue = new PriorityQueue<Pair<Integer, Double>>(RESULT_SIZE, new PairComparator());
		
		double [] validRange = {0.02, 0.04, 0.06, 0.08, 0.1};
		int [] numValid = new int[validRange.length];
		
		//retrieve top-$RESULT_SIZE documents w.r.t. the original input
		for (Entry<Integer, Double> entry : this.sQueryDocMatrix.getRowVector(inputId))
		{
			int docId = entry.getKey();
			
			double text = 0;
			if (AOL)
				text = this.sRawQueryDocMatrix.get(inputId, docId);
			else
			{
				for (String word : inputWords)
				{
					if (this.queryStringToIdMap.containsKey(word))
					{
						int id = this.queryStringToIdMap.get(word);
						text += this.sRawQueryDocMatrix.get(id, docId);					
					}
				}				
			}
			
			docQueue.add(new Pair<Integer, Double>(docId, text));
			if (docQueue.size() > RESULT_SIZE)
				docQueue.poll();

			if (weighter.getDocDist(docId) <= 0.1)
			{
				nearbyTopDocQueue.add(new Pair<Integer, Double>(docId, text));
				if (nearbyTopDocQueue.size() > RESULT_SIZE)
					nearbyTopDocQueue.poll();
			}
		}
		
		PrintWriter userStudyWriter = new PrintWriter(new FileWriter(this.resultPathPrefix + "user_study" + beta + ".txt", true));
		userStudyWriter.println(this.idToStringArray[inputId]);
//		userStudyWriter.println(queryLoc.getLocx() + "\t" + queryLoc.getLocy());
//		while (nearbyTopDocQueue.isEmpty() == false)
//		{
//			Pair<Integer, Double> doc = nearbyTopDocQueue.poll();
//			int docId = doc.left;
//			String text = this.docToTextArray[docId];
//			userStudyWriter.print(text + "\t");
//		}
//		userStudyWriter.println();
//		userStudyWriter.println();

		List<Double> originalScore = new ArrayList<Double>();
		List<Integer> originalResult = new ArrayList<Integer>();
		Set<String> originalWord = new HashSet<String>();
		while (docQueue.isEmpty() == false)
		{
			Pair<Integer, Double> doc = docQueue.poll();
			int docId = doc.left;
			double score = doc.right;
				
			originalScore.add(score);

			originalResult.add(docId);
			
			if (AOL == false)
			{
				String text = this.docToTextArray[docId];
				for (String word : text.split(" "))
					originalWord.add(word);
			}
		}
		
		List<Integer> resultList = new ArrayList<Integer>();
        double tfidfScoreStat = 0;
        double cosScoreStat = 0;
		PrintWriter writer = new PrintWriter(new FileWriter(resultPath, true));
		writer.println(inputId + ": " + this.idToStringArray[inputId] + " (" + queryLoc.getLocx() + ", " + queryLoc.getLocy() + ")");
		writer.println("alpha = " + alpha + " beta = " + beta + " gamma = " + gamma + " epsilon = " + threshold);

		PrintWriter phraseWriter = new PrintWriter(new FileWriter(this.resultPathPrefix + "phrase" + beta + ".txt", true));
		phraseWriter.println(this.idToStringArray[inputId]);
		phraseWriter.println();
		
		
        double normSuggests = 1;
        for (int i = 2; i <= k; ++i)
        	normSuggests += 1.0 / Math.log(i);
		
		double expectedQuery = 0;
		double expectedDoc = 0;

		sortedResult.poll();
		for (int i = 1; i < Math.min(sortedResult.size(), k + 1); ++i)
		{
			SortedMapTreeSet<Integer, Double>.KeyValuePair top = sortedResult.poll();
			int phraseId = top.key;
			if (phraseId == query.expectedQueryId)
				expectedQuery += 1;
			

			String phrase = idToStringArray[phraseId];
			phraseWriter.println(phrase);
			
			userStudyWriter.println(phrase);
			
			double score = top.value;
			String [] suggestWords = phrase.split(" ");
			resultList.add(phraseId);

			SVector docVector = sQueryDocMatrix.getRowVector(phraseId);
			//for each document relevant to suggested query
			for (Entry<Integer, Double> docEntry : docVector)
			{
				int docId = docEntry.getKey();
				
				double dist = weighter.getDocDist(docId);
				
				//check the distance of the document
				for (int jj = validRange.length - 1; jj >= 0; --jj)
				{
					//largest to smallest
					if (dist <= validRange[jj])
						numValid[jj]++;
					else
						break;
				}
				
				//score of this document w.r.t. the suggested keyword
				double text = 0;
				if (AOL)
					text = this.sRawQueryDocMatrix.get(phraseId, docId);
				else
				{
					for (String word : suggestWords)
					{
						if (this.queryStringToIdMap.containsKey(word))
						{
							int id = this.queryStringToIdMap.get(word);
							text += this.sRawQueryDocMatrix.get(id, docId);						
						}
					}					
				}
			
				//top documents w.r.t. the suggested keyword
				docQueue.add(new Pair<Integer, Double>(docId, text));
				if (docQueue.size() > RESULT_SIZE)
					docQueue.poll();

				if (dist <= 0.1)
				{
					nearbyTopDocQueue.add(new Pair<Integer, Double>(docId, text));
					if (nearbyTopDocQueue.size() > RESULT_SIZE)
						nearbyTopDocQueue.poll();
				}
			}
			
			for (int ii = 0; ii < RESULT_SIZE; ++ii)
			{
				if (nearbyTopDocQueue.isEmpty() == false)
				{
					Pair<Integer, Double> doc = nearbyTopDocQueue.poll();
					int docId = doc.left;
					String text = this.docToTextArray[docId];
					userStudyWriter.println(text);
				}
				else
				{
					userStudyWriter.println();
				}
			}

//			SVector orgVec = new SVector();
//			SVector newVec = new SVector();
			
			Set<Integer> newResult = new HashSet<Integer>();
			int rank = docQueue.size();
			double cosScore = 0;
			//for each top document
			while (docQueue.isEmpty() == false)
			{
				Pair<Integer, Double> doc = docQueue.poll();
				int docId = doc.left;
								
				if (docId == query.expectedDocId)
					expectedDoc += 1;
				
				if (!AOL && rank <= originalResult.size())
				{
					Set<String> wordSet = new HashSet<String>();
					newResult.add(docId);
					String text = this.docToTextArray[docId];
					for (String word : text.split(" "))
						wordSet.add(word);
					wordSet.addAll(originalWord);
					
					Map<String, Integer> wordIdMap = new HashMap<String, Integer>();
					int ii = 0;
					for (String word : wordSet)
						wordIdMap.put(word, ii++);
					
					SVector orgVec = new SVector();
					SVector newVec = new SVector();
					
					String originalString = this.docToTextArray[originalResult.get(rank - 1)];					
					for (String word : originalString.split(" "))
						orgVec.add(wordIdMap.get(word), 1);

					String newString = this.docToTextArray[docId];
					for (String word : newString.split(" "))
						newVec.add(wordIdMap.get(word), 1);
					
					double cosSim = orgVec.cosineSimilarity(newVec);
					if (rank > 1)
						cosSim /= Math.log(rank);
					cosScore += cosSim;
				}
				
				rank--;
			}
			
			double normDoc = 1;
			for (int ii = 2; ii < docQueue.size(); ++ii)
				normDoc += 1.0 / Math.log(ii);

			//cosScore = orgVec.cosineSimilarity(newVec);
			cosScore /= normDoc;
			
			for (Entry<Integer, Double> docEntry : docVector)
			{
				int docId = docEntry.getKey();
				
				double dist = weighter.getDocDist(docId);
				
				if (dist > RANGE)
					continue;
				
				double text = 0;
				if (AOL)
					text = this.sRawQueryDocMatrix.get(phraseId, docId);
				else
				{
					for (String word : suggestWords)
					{
						if (this.queryStringToIdMap.containsKey(word))
						{
							int id = this.queryStringToIdMap.get(word);
							text += this.sRawQueryDocMatrix.get(id, docId);						
						}
					}					
				}
			
				docQueue.add(new Pair<Integer, Double>(docId, text));
				if (docQueue.size() > RESULT_SIZE)
					docQueue.poll();
			}
			
			normDoc = 1;
			for (int ii = 2; ii < docQueue.size(); ++ii)
				normDoc += 1.0 / Math.log(ii);

			double tfidfScore = 0;
			rank = docQueue.size();
			while (docQueue.isEmpty() == false)
			{
				Pair<Integer, Double> doc = docQueue.poll();
				
				//double tfidfSuggest = this.sRawQueryDocMatrix.get(phraseId, doc.left);//doc.right;
				
				double tfidfSuggest = 0;
				for (String word : inputWords)
				{
					if (this.queryStringToIdMap.containsKey(word))
					{
						int id = this.queryStringToIdMap.get(word);
						tfidfSuggest += this.sRawQueryDocMatrix.get(id, doc.left);
					}
				}
				
				double ratio = tfidfSuggest;
				if (rank <= originalScore.size())
					ratio /= originalScore.get(rank - 1);
				
				if (rank > 1)
					ratio /= Math.log(rank);
				tfidfScore += ratio;
				
				rank--;
			}
			
			tfidfScore /= normDoc;
			
			if (i > 1)
			{
				tfidfScore /= Math.log(i);
				cosScore /= Math.log(i);
			}
			
			tfidfScoreStat += tfidfScore;
			cosScoreStat += cosScore;
			
			writer.println("[" + i + "] " + phrase + "\t" + score + "\t" + tfidfScore + "\t" + cosScore);
			
//			if (this.docToTextArray != null)
//			{
//				for (int ii = docList.size() - 1; ii >= 0; --ii)
//				{
//					int dId = docList.get(ii);
//					writer.println("\t" + this.docToTextArray[dId]);
//					writer.println("\t" + this.docToLocArray[dId].getLocx() + "\t" + this.docToLocArray[dId].getLocy());
//				}
//			}
		}

		phraseWriter.println();
		phraseWriter.close();
		
		userStudyWriter.println("------------------------");
		userStudyWriter.println();
		userStudyWriter.close();
		
		tfidfScoreStat /= normSuggests;
		cosScoreStat /= normSuggests;
		
//		if (sortedCluster != null)
//		{
//			double distCost = 0, accuCost = 0;
//			for (int i = 0; i != numCluster; ++i)
//			{
//				QueryCluster cluster = sortedCluster.get(i);
//				writer.printf("%d(%.4f) ", cluster.statNumDist, cluster.statInk);
//				distCost += cluster.distCost;
//				accuCost += cluster.accuCost;
//			}
//			writer.println();
//			writer.println("DistCost = " + distCost);
//			writer.println("AccuCost = " + accuCost);
//		}
//		
//		if (sortedCell != null)
//		{
//			double distCost = 0, accuCost = 0;
//			for (int i = 0; i != numCluster; ++i)
//			{
//				DocCluster cell = sortedCell.get(i);
//				writer.printf("%d(%.4f) ", cell.statNumDist, cell.statInk);
//				distCost += cell.distCost;
//				accuCost += cell.accuCost;
//			}
//			writer.println();
//			writer.println("DistCost = " + distCost);
//			writer.println("AccuCost = " + accuCost);
//		}
		
		writer.println("textScore = " + tfidfScoreStat + "\tdocSim = " + cosScoreStat);
		for (int jj = 0; jj < validRange.length; ++jj)
			writer.println("numValid(" + validRange[jj] + ") = " + numValid[jj]);
		writer.println("rInk = " + rInk);
		writer.println("iter = " + iter);
		writer.println("cost = " + cost);
		writer.println("****************************");
		writer.println();
		writer.close();

		double similarityToTruth = 0;
		double misMatch = 0;
		if (truthList.isEmpty())
		{
			truthList.addAll(resultList);
			similarityToTruth = 1;
			misMatch = 0;
		}
		else
		{
			misMatch = Tools.countMismatch(resultList, truthList);
			//textScoreStat = Tools.jaccardSimilarity(resultList, truthList);
			//textScoreStat = Tools.getNDCG(resultList, truthList);
			similarityToTruth = Tools.averagePrecision(resultList, truthList);
		}

		ResultStat stat = new ResultStat();
		stat.add(cost);
		stat.add(iter);
		stat.add(tfidfScoreStat);
		stat.add(similarityToTruth);
		stat.add(misMatch);
		stat.add(cosScoreStat);
		stat.add(rInk);
		for (double num : numValid)
			stat.add(num / k);
		stat.add(expectedQuery);
		stat.add(expectedDoc);
		
		return stat;
	}
}
