package sugg;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.PriorityQueue;

import edu.mit.jwi.Dictionary;
import edu.mit.jwi.IDictionary;
import edu.mit.jwi.morph.IStemmer;
import edu.mit.jwi.morph.WordnetStemmer;
import sugg.invertedindex.InvertedIndex;
import sugg.invertedindex.PhraseInfo;
import sugg.invertedindex.Posting;
import sugg.matrix.SMatrix;
import sugg.matrix.SVector;
import sugg.objects.Location;
import sugg.tools.Timer;
import sugg.tools.Tools;

public class DataGenerator{
	
	public static class Cell implements Comparable<Cell>{
		public int id;
		public int row;
		public int col;
		public List<Integer> docList;
		public List<Integer> neighbors;
		public Location loc;
		private int totalDoc;
		
		public Cell(int id, int N, int totalDoc)
		{
			this.id = id;
			this.row = this.id / N;
			this.col = this.id % N;
			this.loc = new Location((double)(0.5 + this.row) / N, (double)(0.5 + this.col) / N);
			this.totalDoc = totalDoc;
			
			this.docList = new ArrayList<Integer>();
			this.neighbors = new ArrayList<Integer>();
			
			if (this.row != 0)
				this.neighbors.add(id - N);
			if (this.col != 0)
				this.neighbors.add(id - 1);
			if (this.row != N - 1)
				this.neighbors.add(id + N);
			if (this.col != N - 1)
				this.neighbors.add(id + 1);

			if (this.row != 0 && this.col != 0)
				this.neighbors.add(id - N - 1);
			if (this.row != 0 && this.col != N - 1)
				this.neighbors.add(id - N + 1);
			if (this.row != N - 1 && this.col != 0)
				this.neighbors.add(id + N - 1);
			if (this.row != N - 1 && this.col != N - 1)
				this.neighbors.add(id + N + 1);
		}

		public double getWeight()
		{
			if (this.docList.size() == 0)
				return 0.0;
			if (this.totalDoc == 0)
				Tools.printError("Never gonna happen: totalDoc == 0 & docList != 0");
			return ((double)this.docList.size()) / this.totalDoc;
		}

		@Override
		public int compareTo(Cell other) {
			//sort in ascending order
			if (this.getWeight() > other.getWeight())
				return 1;
			else
				return -1;
		}
	}
	
	private String inputType;
	
	public int numQuery;
	public int numDoc;	
	public long numEdge;

	private Location [] oldDocToLocArray;
	private String [] oldDocToStringArray;
	private InvertedIndex invertedIndex;
	private SMatrix sQueryDocMatrix;
	
	private String outputPathPrefix;
	private String infoFilePath;
	private String queryStringFilePath;
	private String docLocationFilePath;
	private String docTextFilePath;
	
	private Indexer indexer;
	
	String [] queryIdToStringArray;
	Map<String, Integer> queryStringToIdMap;
	Map<String, List<Integer>> queryStringToClusterMap;
	Location [] docToLocArray;
	
	public DataGenerator(String inputType) throws IOException
	{
		this.inputType = inputType;
		
		this.outputPathPrefix = "output/" + inputType + "/";
		Tools.mkdir(this.outputPathPrefix);

		this.infoFilePath = this.outputPathPrefix + "info.txt";
		this.queryStringFilePath = this.outputPathPrefix + "querystring.txt";
		this.docLocationFilePath = this.outputPathPrefix + "docloc.txt";
		this.docTextFilePath = this.outputPathPrefix + "doctext.txt";
		
		Tools.mkdir(this.outputPathPrefix + "matrix/");
		
		this.indexer = new Indexer(Main.sourcefile, Main.indexfile, Main.inputType, 
				Main.num, Main.buildLuceneIndex, Main.buildInvertedIndex, Main.threshold, Main.maxGram);
		this.invertedIndex = indexer.getInvertedIndex();
		this.oldDocToLocArray = indexer.getDocToLocArray();
		this.oldDocToStringArray = indexer.getDocToTextArray();
	}

	public void generate(int cellN) throws IOException
	{
		this.numDoc = indexer.numDocs;
		this.numQuery = this.invertedIndex.size();
		
		Timer timer = new Timer();
		
		SMatrix sQueryDocMatrix = new SMatrix(this.numQuery, this.numDoc);
		
		double interval = 1.0 / cellN;
		int [] docToCellArray = new int[this.numDoc];
		int [] cellSize = new int[cellN * cellN];
		for (int i = 0; i != this.numDoc; ++i)
		{
			int cellId = Tools.locToCellID(this.oldDocToLocArray[i], cellN, interval);
			docToCellArray[i] = cellId;
			cellSize[cellId]++;
		}
		
		String dictPath = "data/dict";
		URL url = new URL("file", null, dictPath);
		IDictionary dict = new Dictionary(url);
		dict.open();
		
		IStemmer stemmer = new WordnetStemmer(dict);
		
		double cost = 0;
		int phraseId = 0;
		int iter = 0;
		List<String> phraseList = new ArrayList<String>();
		Map<String, List<Integer>> phraseClusterMap = new HashMap<String, List<Integer>>();
		for (String phrase : this.invertedIndex.keySet())
		{
			if (iter % 10000 == 0)
				Tools.printProgress(iter, this.numQuery, cost);
			iter++;
			
			timer.start();
			
			if (Tools.filter(phrase, dict, stemmer))
				continue;

			PhraseInfo phraseInfo = this.invertedIndex.get(phrase);

			if (this.inputType.contains("tripadvisor") && (phraseInfo.postList.size() <= 5 || phraseInfo.postList.size() >= 10000))
				continue;
			
			Cell [] cellArray = new Cell[cellN * cellN];
			for (int i = 0; i != cellN * cellN; ++i)
				cellArray[i] = new Cell(i, cellN, cellSize[i]);

			for (Posting posting : phraseInfo.postList)
			{
				int docId = posting.getId();
				int cellId = docToCellArray[docId];
				cellArray[cellId].docList.add(docId);
			}
			
			int c = 5;	//5 cluster cells
			PriorityQueue<Cell> clusetCellHeap = new PriorityQueue<Cell>();
			for (int i = 0; i != cellN * cellN; ++i)
			{
				clusetCellHeap.add(cellArray[i]);
				if (clusetCellHeap.size() > c)
					clusetCellHeap.poll();
			}
			
			List<Integer> clusterList = new ArrayList<Integer>();
			while (clusetCellHeap.isEmpty() == false)
			{
				Cell cell = clusetCellHeap.peek();
				clusetCellHeap.poll();
				if (cell.getWeight() == 0)
					continue;
				clusterList.add(cell.id);
			}
			phraseClusterMap.put(phrase, clusterList);
			
			for (Posting posting : phraseInfo.postList)
			{
				int docId = posting.getId();
				int tf = posting.getTf();
				
				double score = Math.log(tf + 1) * phraseInfo.idf;	//use tf idf only
				score /= this.indexer.getAvgGramFreq()[phrase.split(" ").length - 1];
				sQueryDocMatrix.set(phraseId, docId, score);
			}
			
			phraseList.add(phrase);
			phraseId++;
			
			cost += timer.stop();
		}
		System.out.println();
		dict.close();
		assert(phraseId == phraseList.size());

		System.out.println("Compressing the matrix ...");
		List<Integer> newQueryId_oldQueryId = new ArrayList<Integer>();
		List<String> nonEmptyPhraseList = new ArrayList<String>();
		List<Integer> newDocId_oldDocId = new ArrayList<Integer>();
		List<Location> nonEmptyLocationList = new ArrayList<Location>();
		List<String> nonEmptyTextList = new ArrayList<String>();
		for (int i = 0; i != sQueryDocMatrix.numRows(); ++i)
		{
			SVector vec = sQueryDocMatrix.getRowVector(i);
			if (vec.normL1() > Tools.EP)	//vec.normL1() > 0
			{
				newQueryId_oldQueryId.add(i);
				nonEmptyPhraseList.add(phraseList.get(i));
			}
		}
		
		SMatrix sDocQueryMatrix = sQueryDocMatrix.transpose();
		for (int i = 0; i != sDocQueryMatrix.numRows(); ++i)
		{
			SVector vec = sDocQueryMatrix.getRowVector(i);
			if (vec.normL1() > Tools.EP)	//vec.normL1() > 0
			{
				newDocId_oldDocId.add(i);
				nonEmptyLocationList.add(this.oldDocToLocArray[i]);
				nonEmptyTextList.add(this.oldDocToStringArray[i]);
			}
		}
		
		Map<Integer, Integer> oldDocId_newDocId = new HashMap<Integer, Integer>();
		for (int i = 0; i < newDocId_oldDocId.size(); ++i)
		{
			int newDocId = i;
			int oldDocId = newDocId_oldDocId.get(i);
			oldDocId_newDocId.put(oldDocId, newDocId);
		}
		
		SMatrix compressedMatrix = new SMatrix(newQueryId_oldQueryId.size(), newDocId_oldDocId.size());
		for (int newQueryId = 0; newQueryId < newQueryId_oldQueryId.size(); ++newQueryId)
		{
			if (newQueryId % 1000 == 0)
				System.out.printf("\rCompressing %d(%d)", newQueryId, newQueryId_oldQueryId.size());
			int queryId = newQueryId_oldQueryId.get(newQueryId);		//old query id
			SVector vec = sQueryDocMatrix.getRowVector(queryId);	//old query-doc vector
			for (Entry<Integer, Double> docEntry : vec)
			{
				int oldDocId = docEntry.getKey();
				if (oldDocId_newDocId.containsKey(oldDocId))
				{
					int newDocId = oldDocId_newDocId.get(oldDocId);
					double val = docEntry.getValue();
					compressedMatrix.set(newQueryId, newDocId, val);
				}
			}
		}
		
		System.out.println("numQuery = " + compressedMatrix.numRows());
		System.out.println("numDoc = " + compressedMatrix.numColumns());
		
		this.sQueryDocMatrix = compressedMatrix;
		this.numQuery = compressedMatrix.numRows();
		this.numDoc = compressedMatrix.numColumns();
		
		this.output(nonEmptyPhraseList, nonEmptyLocationList, nonEmptyTextList,
				newDocId_oldDocId, phraseClusterMap,
				this.sQueryDocMatrix);
	}
	
	private void output(List<String> phraseList, List<Location> locationList,
			List<String> textList,
			List<Integer> newDocId_oldDocId,
			Map<String, List<Integer>> phraseClusterMap,
			SMatrix sQueryDocMatrix) throws IOException
	{
		PrintWriter writer = new PrintWriter(this.queryStringFilePath);
		for (int i = 0; i != this.numQuery; ++i)
		{
			String phrase = phraseList.get(i);
			writer.println(phrase);
			for (int cellId : phraseClusterMap.get(phrase))
				writer.print(cellId + " ");
			writer.println();
		}
		writer.close();
		
		writer = new PrintWriter(this.docLocationFilePath);
		for (int j = 0; j != this.numDoc; ++j)
			writer.println(locationList.get(j).getLocx() + "\t" + locationList.get(j).getLocy());
		writer.close();
		
		writer = new PrintWriter(this.docTextFilePath);
		for (int j = 0; j != this.numDoc; ++j)
			writer.println(textList.get(j));
		writer.close();
		
		this.numEdge = sQueryDocMatrix.numEntries();
		
		writer = new PrintWriter(this.infoFilePath);
		writer.println("numQuery = " + this.numQuery);
		writer.println("numEdge = " + this.numEdge);
		writer.println("numDoc = " + this.numDoc);
		writer.close();
		
		String rawQueryDocMatrixPath = this.outputPathPrefix + "matrix/query_doc.raw.matrix";
		sQueryDocMatrix.write(new File(rawQueryDocMatrixPath), false);
	}
}
