package sugg;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Scanner;

import org.apache.jdbm.BTree;
import org.apache.jdbm.RecordHeader;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.DoubleField;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.FieldType;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.DocsEnum;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.IndexWriterConfig;
import org.apache.lucene.index.MultiFields;
import org.apache.lucene.index.TermsEnum;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.similarities.DefaultSimilarity;
import org.apache.lucene.search.similarities.TFIDFSimilarity;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.util.BytesRef;
import org.apache.lucene.util.Version;

import sugg.invertedindex.InvertedIndex;
import sugg.invertedindex.InvertedIndexIsolated;
import sugg.invertedindex.InvertedIndexPacked;
import sugg.invertedindex.PhraseInfo;
import sugg.invertedindex.Posting;
import sugg.objects.Location;
import sugg.objects.Pair;
import sugg.tools.Timer;
import sugg.tools.Tools;

public class Indexer
{
	private int maxGram;
	private double [] avgiGramFreq;
	private Directory directory;
	private Analyzer analyzer;
	private IndexReader indexReader;
	public String indexPathPrefix;
	private String outputPathPrefix;
	private String locationFilename;
	private String textFilename;
	private String field;
	
	public String inputType;
	public int numDocs;
	private Location [] docToLocArray;
	private String [] docToTextArray;
	public double minLocx = Double.MAX_VALUE;
	public double maxLocx = -Double.MAX_VALUE;
	public double minLocy = Double.MAX_VALUE;
	public double maxLocy = -Double.MAX_VALUE;
	
	private int threshold;
	private InvertedIndex invertedIndex;
	private int maxLength = 0;
	private double maxTfIdf = -Double.MAX_VALUE;
	private double minTfIdf = Double.MAX_VALUE;
	private int maxTf = Integer.MIN_VALUE;
	private int minTf = Integer.MAX_VALUE;
	private float maxIdf = -Float.MAX_VALUE;
	private float minIdf = Float.MAX_VALUE;
	
		
	public Indexer (String filename, String indexfile, String inputType, int num, 
			boolean buildIndex, boolean buildInvertedIndex,
			int threshold, int maxGram)
					throws IOException
	{
		System.out.println("*****************************************************");
		
		this.inputType = inputType;
		this.indexPathPrefix = "index/" + inputType.replace("/synthetic", "") + "/";
		Tools.mkdir(this.indexPathPrefix);
		this.outputPathPrefix = "output/" + inputType + "/";
		Tools.mkdir(this.outputPathPrefix);
		
		this.locationFilename = indexPathPrefix + "location.txt";
		this.textFilename = indexPathPrefix + "text.txt";
		this.field = inputType.startsWith("twitter") ? "text" : "reviews";
		
		this.maxGram = maxGram;
		this.analyzer = new MyAnalyzer(Version.LUCENE_46, maxGram);
		this.directory = FSDirectory.open(new File(indexfile));
		this.threshold = threshold;
		
		if (buildIndex == true)
		{
			System.out.print("Build new lucene index? Type \"YES\" to confirm: ");
			Scanner in = new Scanner(System.in);
			String confirm = in.next();
			in.close();
			if (confirm.equals("YES") == false)
			{
				this.close();
				System.exit(0);
			}
		}
		
		buildIndex = (buildIndex == true || DirectoryReader.indexExists(directory) == false);
		
		if(buildIndex == true)
		{
			if (inputType.startsWith("twitter"))
			{
				this.buildLuceneIndexTwitter(filename, num);
			}
		}

		this.indexReader = DirectoryReader.open(this.directory);
		
		this.numDocs = indexReader.numDocs();

		this.getDocuments();
		this.getInvertedIndex(0, buildInvertedIndex, false);
	}
	
	private void getInvertedIndex (int hrN, boolean buildInvertedIndex, boolean useIsolated) throws IOException
	{
		Timer timer = new Timer();
		
		String invertedIndexFolderPath = this.indexPathPrefix;
		
		if (useIsolated)
			invertedIndexFolderPath += "inverted_index_v2/";
		else
			invertedIndexFolderPath += "inverted_index_v1/";
		
		String treeFilePrefix = invertedIndexFolderPath + "inverted_index_M" + BTree.DEFAULT_SIZE +  "_HN" + hrN;
		
		String infofile = treeFilePrefix + ".info.txt";
		
		boolean fileExists = new File(treeFilePrefix + ".db.d.0").isFile();
		
		String treeName = "inverted_index_HN" + hrN;
		
		if (fileExists == true && buildInvertedIndex == false)
		{
			if (useIsolated)
				this.invertedIndex = new InvertedIndexIsolated(treeFilePrefix, treeName, false);
			else
				this.invertedIndex = new InvertedIndexPacked(treeFilePrefix, treeName, false);
			
			Scanner scanner = new Scanner(new File(infofile));
			this.maxLength = scanner.nextInt();
			this.maxTfIdf = scanner.nextDouble();
			this.minTfIdf = scanner.nextDouble();
			this.maxTf = scanner.nextInt();
			this.minTf = scanner.nextInt();
			this.maxIdf = scanner.nextFloat();
			this.minIdf = scanner.nextFloat();
			this.maxGram = scanner.nextInt();
			
			this.avgiGramFreq = new double[this.maxGram];
			for (int i = 0; i != this.maxGram; ++i)
			{
				this.avgiGramFreq[i] = scanner.nextDouble();
				this.avgiGramFreq[i] = Math.log(this.avgiGramFreq[i]);
			}
			scanner.close();
		}
		else
		{
			File invertedIndexFolder = new File(invertedIndexFolderPath);
            if(invertedIndexFolder.isDirectory()){
            	for (String file : invertedIndexFolder.list())
            		new File(file).delete();
            }
            else
            	invertedIndexFolder.mkdirs();

			//inverted file doesn't exist, create one
			System.out.println("----Building Inverted Index TreeMap ...");
			timer.start();

			if (useIsolated)
				this.invertedIndex = new InvertedIndexIsolated(treeFilePrefix, treeName, true);
			else
				this.invertedIndex = new InvertedIndexPacked(treeFilePrefix, treeName, true);

			this.avgiGramFreq = new double[this.maxGram];
			int [] gramCount = new int[this.maxGram];
			
			TermsEnum termsEnum = MultiFields.getTerms(this.indexReader, this.field).iterator(null);
			TFIDFSimilarity tfidfSim = new DefaultSimilarity();
			BytesRef bytesRef;
			
			int id = 0;
			while ((bytesRef = termsEnum.next()) != null)
			{
				long tfSum = termsEnum.totalTermFreq();
				if (tfSum < this.threshold)
					continue;
				
				String phrase = bytesRef.utf8ToString();
				if (Tools.filter(phrase))
					continue;
				
				if (phrase.length() > this.maxLength)
					this.maxLength = phrase.length();
				
				PhraseInfo phraseInfo = new PhraseInfo();
				
				phraseInfo.id = id++;
				phraseInfo.tfSum = tfSum;
				phraseInfo.idf = tfidfSim.idf(termsEnum.docFreq(), this.indexReader.numDocs());

				int gramOrder = phrase.split(" ").length;
				gramCount[gramOrder - 1]++;
				this.avgiGramFreq[gramOrder - 1] += phraseInfo.tfSum;

				Map<Integer, Integer> postingMap = new HashMap<Integer, Integer>();
			    DocsEnum docsEnum = MultiFields.getTermDocsEnum(this.indexReader, MultiFields.getLiveDocs(this.indexReader), this.field, bytesRef);
			    while(docsEnum.nextDoc() != DocsEnum.NO_MORE_DOCS)
			    {
			    	int docID = docsEnum.docID();
			    	int tf = docsEnum.freq();
			    	postingMap.put(docID, tf);
			    }

			    for (Entry<Integer, Integer> tfEntry : postingMap.entrySet())
			    {
			    	Posting posting = new Posting(tfEntry.getKey(), tfEntry.getValue());
			    	phraseInfo.postList.add(posting);
			    }
			    
			    if (phraseInfo.getSizeInByte() > 0.7 * RecordHeader.MAX_RECORD_SIZE)
			    {
			    	System.out.println("Dude this is huuuge: " + phrase + " " + phraseInfo.postList.size());
			    	continue;
			    }
    
			    phraseInfo.sortPostList();
				float idf = phraseInfo.idf;
			    
			    try {
				    this.invertedIndex.put(phrase, phraseInfo);
				    
					Integer maxTf = phraseInfo.postList.get(0).getTf();
				    Integer minTf = phraseInfo.postList.get(phraseInfo.postList.size() - 1).getTf();
				    
				    this.maxTfIdf = maxTf * idf > this.maxTfIdf ? maxTf * idf : this.maxTfIdf;
				    this.minTfIdf = minTf * idf < this.minTfIdf ? minTf * idf : this.minTfIdf;
				    
				    this.maxTf = maxTf > this.maxTf ? maxTf : this.maxTf;
				    this.minTf = minTf < this.minTf ? minTf : this.minTf;
				    
				    this.maxIdf = idf > this.maxIdf ? idf : this.maxIdf;
				    this.minIdf = idf < this.minIdf ? idf : this.minIdf;
			    }
			    catch (Exception e)
			    {
			    	System.out.println("Exception: " + phraseInfo.getSizeInByte());
			    }			    
			}

			PrintWriter writer = new PrintWriter(new File(infofile));
			writer.println(this.maxLength);
			writer.println(maxTfIdf);
			writer.println(minTfIdf);
			writer.println(maxTf);
			writer.println(minTf);
			writer.println(maxIdf);
			writer.println(minIdf);
			
			//not necessarily useful
			writer.println(this.maxGram);
			for (int i = 0; i < this.maxGram; ++i)
			{
				avgiGramFreq[i] = this.avgiGramFreq[i] / gramCount[i];
				writer.println(avgiGramFreq[i]);
				System.out.printf("Average %d-gram frequency = %f (%f)\n", i + 1, avgiGramFreq[i], Math.log(avgiGramFreq[i]));
				avgiGramFreq[i] = Math.log(avgiGramFreq[i]);
			}
			writer.println(this.invertedIndex.size());

			System.out.println("Inverted Index TreeMap built in " + timer.stop() + " seconds.");
			
			writer.close();
		}
	}
	
	public void getDocRelations(int m) throws IOException
	{	
		String docRelationFilePath = this.indexPathPrefix + "doclist.txt";
		File docRelationFile = new File(docRelationFilePath);
		if (docRelationFile.isFile() == false)
		{
			PrintWriter writer = new PrintWriter(docRelationFile);
			
			IndexSearcher indexSearcher = new IndexSearcher(this.indexReader);
			QueryParser parser = new QueryParser(Version.LUCENE_46, this.field, this.analyzer);
			
			for (int i = 0; i != this.numDocs; ++i)
			{
				if (i % 10000 == 0)
					System.out.printf("\r%d", i);
				
				List<Pair<Integer, Float>> docList = new ArrayList<Pair<Integer, Float>>();
				//System.out.print(i + ": ");
				Document doc = indexSearcher.doc(i);
				
				String text = doc.getField(this.field).stringValue();
				Query query = null;
				try {
					query = parser.parse(text);
				} catch (Exception e) {
					writer.println();
					//System.out.println();
					continue;
				}
				ScoreDoc [] hits = indexSearcher.search(query, m + 1).scoreDocs;
				for (int j = 1; j < hits.length; ++j)
				{
					ScoreDoc hit = hits[j];
					if (indexSearcher.doc(hit.doc).getField(this.field).stringValue().equals(text))
						continue;

					if (hit.score > 0)
					{
						docList.add(new Pair<Integer, Float>(hit.doc, hit.score));
						//System.out.print(hit.doc + " " + hit.score + " ");
						writer.print(hit.doc + " " + hit.score + " ");						
					}
				}
				writer.println();
				//System.out.println();
			}
			writer.close();
		}
	}
	
	public InvertedIndex getInvertedIndex()
	{
		return this.invertedIndex;
	}
	
	public double [] getAvgGramFreq()
	{
		return this.avgiGramFreq;
	}
	
	public Location [] getDocToLocArray()
	{
		return this.docToLocArray;
	}
	
	public String [] getDocToTextArray()
	{
		return this.docToTextArray;
	}
	
	private void getDocuments() throws IOException
	{
		File locationFile = new File(this.locationFilename);
		File textFile = new File(this.textFilename);
		if (locationFile.isFile())
		{
			Scanner locScanner = new Scanner(locationFile);
			Scanner textScanner = new Scanner(textFile);
			
			this.numDocs = locScanner.nextInt();
			if (this.numDocs != indexReader.numDocs())
			{
				System.out.println("**ERROR**: Unmatched number of documents.");
			}

			this.docToLocArray = new Location[this.numDocs];
			this.docToTextArray = new String[this.numDocs];
			
			for (int docID = 0; docID < this.numDocs; ++docID)
			{
				double locx = locScanner.nextDouble();
				double locy = locScanner.nextDouble();
				this.docToLocArray[docID] = new Location(locx, locy);
				
				this.docToTextArray[docID] = textScanner.nextLine();
			}
			textScanner.close();
			
			this.minLocx = locScanner.nextDouble();
			this.maxLocx = locScanner.nextDouble();
			this.minLocy = locScanner.nextDouble();
			this.maxLocy = locScanner.nextDouble();
			locScanner.close();
		}
		else
		{
			System.out.println("----Building location & text information ...");
			
			PrintWriter textWriter = new PrintWriter(textFile);
			PrintWriter locationWriter = new PrintWriter(locationFile);
			
			this.numDocs = indexReader.numDocs();
			locationWriter.println(this.numDocs);

			this.docToLocArray = new Location[this.numDocs];
			this.docToTextArray = new String[this.numDocs];
			for (int docID = 0; docID < numDocs; ++docID)
			{
				Document doc = this.indexReader.document(docID);
				double locx = doc.getField("locx").numericValue().doubleValue();
				this.minLocx = locx < this.minLocx ? locx : this.minLocx;
				this.maxLocx = locx > this.maxLocx ? locx : this.maxLocx;
				
				double locy = doc.getField("locy").numericValue().doubleValue();
				this.minLocy = locy < this.minLocy ? locy : this.minLocy;
				this.maxLocy = locy > this.maxLocy ? locy : this.maxLocy;

				this.docToLocArray[docID] = new Location(locx, locy);
				
				this.docToTextArray[docID] = doc.getField(this.field).stringValue();
			}
			
			for (int docID = 0; docID < numDocs; ++docID)
			{
				double locx = this.docToLocArray[docID].getLocx();
				double locy = this.docToLocArray[docID].getLocy();
				locx = Tools.getNormalization(locx, this.minLocx, this.maxLocx);
				locy = Tools.getNormalization(locy, this.minLocy, this.maxLocy);
				this.docToLocArray[docID].setLocx(locx);
				this.docToLocArray[docID].setLocy(locy);
				
				locationWriter.println(locx + "\t" + locy);
				
				textWriter.println(this.docToTextArray[docID]);
			}
			
			locationWriter.println(this.minLocx);
			locationWriter.println(this.maxLocx);
			locationWriter.println(this.minLocy);
			locationWriter.println(this.maxLocy);
			System.out.println(this.minLocx + " " + this.maxLocx + " " + this.minLocy + " " + this.maxLocy);
			locationWriter.close();
			textWriter.close();
		}
	}
	
	private void buildLuceneIndexTwitter(String filename, int num) throws IOException
	{
		System.out.println("----Building Lucene Index for " + filename + "...");
		Timer timer = new Timer();
		timer.start();
		
		//this.analyzer = new StandardAnalyzer(Version.LUCENE_46);
		//this.directory = new RAMDirectory();
		IndexWriterConfig config = new IndexWriterConfig(Version.LUCENE_46, this.analyzer);
		IndexWriter indexWriter = new IndexWriter(directory, config);
		indexWriter.deleteAll();
		indexWriter.commit();

		int minDocLength = Integer.MAX_VALUE;
		int maxDocLength = Integer.MIN_VALUE;
		double avgDocLength = 0;

		double minLat = 29;
		double maxLat = 49;
		double minLong = -125;
		double maxLong = -70;
		
//		if (filename.contains("new_york"))
//		{
//			minLat = 40.4;
//			maxLat = 41;
//			minLong = -74;
//			maxLong = -73;
//		}
		
		int count = 0;
		Scanner scanner = new Scanner(new File(filename));
		while (true)
		{
			if (scanner.hasNextLine() == false)
				break;
			
			//scanner.nextLine();	//skip the id line
			String text = scanner.nextLine();
			String locString = scanner.nextLine();
			double locx = Double.valueOf(locString.split("\t")[0]);
			double locy = Double.valueOf(locString.split("\t")[1]);
			
			if (locx < minLat || locx > maxLat || locy < minLong || locy > maxLong)
				continue;

			int length = text.split(" ").length;
			minDocLength = length < minDocLength ? length : minDocLength;
			maxDocLength = length > maxDocLength ? length : maxDocLength;
			avgDocLength += length;
			
			FieldType vecTextField = new FieldType();
			vecTextField.setStoreTermVectors(true);
			vecTextField.setStoreTermVectorPositions(true);
			vecTextField.setStored(true);
			vecTextField.setIndexed(true);
			vecTextField.setTokenized(true);
			
			Document doc = new Document();
			doc.add(new Field("text", text, vecTextField));
			doc.add(new DoubleField("locx", locx, DoubleField.TYPE_STORED));
			doc.add(new DoubleField("locy", locy, DoubleField.TYPE_STORED));
			indexWriter.addDocument(doc);
			
			count++;
			
			if (num > 0)
			{
				if (count >= num)
					break;
			}
		}
		scanner.close();
		
		avgDocLength /= count;
		System.out.println("minDocLength = " + minDocLength);
		System.out.println("maxDocLength = " + maxDocLength);
		System.out.println("avgDocLength = " + avgDocLength);
		System.out.println("number of tweets = " + count);
		
		indexWriter.close();
		
		double secs = timer.stop();
		System.out.println("----INDEX BUILT (" + secs + ") ----");
	}
	
	public void close() throws IOException
	{
		this.invertedIndex.close();
		this.directory.close();
		this.indexReader.close();
		System.out.println("[CLOSED]");
	}
	
}
