package sugg.cluster;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.Set;

import sugg.matrix.SMatrix;
import sugg.matrix.SVector;
import sugg.objects.Location;
import sugg.tools.Timer;
import sugg.tools.Tools;

public class Clusteror {
	private Location [] docToLocArray;
	private int numQuery;
	private int numDoc;
	private String clusterPrefix;

	private SMatrix sRawQueryClusterMatrix;
	private SMatrix sQueryClusterMatrix;
	private SMatrix sDocClusterMatrix;
	
	public Clusteror(int numQuery, int numDoc, 
			SMatrix sQueryDocMatrix, SMatrix sDocQueryMatrix, 
			SMatrix sRawQueryDocMatrix,
			Location [] docToLocArray, String clusterPrefix)
	{
		this.numQuery = numQuery;
		this.numDoc = numDoc;
		this.docToLocArray = docToLocArray;
		this.clusterPrefix = clusterPrefix;
		
		if (sDocQueryMatrix != null)
		{
			this.sQueryClusterMatrix = sDocQueryMatrix.transpose();
			assert(this.sQueryClusterMatrix.numRows() == numQuery);
		}
		if (sQueryDocMatrix != null)
		{
			this.sDocClusterMatrix = sQueryDocMatrix.transpose();
			assert(this.sDocClusterMatrix.numRows() == numDoc);
		}
		if (sRawQueryDocMatrix != null)
		{
			this.sRawQueryClusterMatrix = sRawQueryDocMatrix;
		}
	}
	
	public int [] getDocClusterArray(int N, int M, String method) throws IOException
	{
		int numCluster = N * N * M;
		if (numCluster == 1)
			return new int[this.numDoc];
		
		if (method.startsWith("Rand"))
			return this.getRandomCluster(this.numDoc, numCluster);
		
		String docClusterFilePath = this.clusterPrefix + method + "_" + numCluster 
				+ "_N" + N + "_M" + M + "_doc_cluster.txt";
		File docClusterFile = new File(docClusterFilePath);
		int [] docToClusterArray = null;
		
		if (docClusterFile.isFile())
			docToClusterArray = this.readClusterArray(this.numDoc, docClusterFile);
		else
			docToClusterArray = this.buildDocClusterArray(N, M, docClusterFile, method);
		
		return docToClusterArray;
	}
	
	public int [] getQueryClusterArray (int N, int M, int [] docClusterArray, String method) throws IOException
	{
		int numCluster = N * N * M;
		if (numCluster == 1)
			return new int[this.numQuery];
		
		if (method.endsWith("Rand"))
			return this.getRandomCluster(this.numQuery, numCluster);
		
		String queryClusterFilePath = this.clusterPrefix + method + "_" + numCluster 
				+ "_N" + N + "_M" + M + "_query_cluster.txt";

		int [] queryClusterArray = null;
		File queryClusterFile = new File(queryClusterFilePath);
		
		if (queryClusterFile.isFile())
			queryClusterArray = this.readClusterArray(this.numQuery, queryClusterFile);
		else
			queryClusterArray = this.sBuildQueryClusterArray_Cell(
						this.numQuery, numCluster, docClusterArray, queryClusterFile);
		
		return queryClusterArray;
	}
	
	public void getClusterSortedMatrix (SMatrix matrix, int [] cluster, int numData, int numCluster, String pathPrefix) throws IOException
	{
		PrintWriter writer = new PrintWriter(new File(pathPrefix + "_" + numData + "_" + numCluster + ".sorted.txt"));
		List<Set<Integer>> clusterList = new ArrayList<Set<Integer>>();
		for (int i = 0; i != numCluster; ++i)
			clusterList.add(new HashSet<Integer>());
		for (int i = 0; i != numData; ++i)
		{
			int clusterId = cluster[i];
			clusterList.get(clusterId).add(i);
		}
		for (Set<Integer> data : clusterList)
		{
			for (int id : data)
			{
				SVector vec = matrix.getRowVector(id);
				for (Entry<Integer, Double> entry : vec)
					writer.println(id + "\t" + entry.getKey() + "\t" + entry.getValue());
			}
		}
		writer.close();
	}
	
	private List<List<Integer>> buildDocCellRand(int numCell)
	{
		List<List<Integer>> cellDocListList = new ArrayList<List<Integer>>();
		for (int i = 0; i != numCell; ++i)
			cellDocListList.add(new ArrayList<Integer>());
		
		for (int i = 0; i != this.numDoc; ++i)
		{
			int cellId = (int) (Math.random() * numCell);
			cellDocListList.get(cellId).add(i);
		}
		
		return cellDocListList;
	}
	
	private List<List<Integer>> buildDocCell(int sqrtN, int numCell)
	{
		List<List<Integer>> cellDocListList = new ArrayList<List<Integer>>();
		for (int i = 0; i != numCell; ++i)
			cellDocListList.add(new ArrayList<Integer>());
		for (int i = 0; i != this.numDoc; ++i)
		{
			int cellId = Tools.locToCellID(this.docToLocArray[i], sqrtN, 1.0/sqrtN);
			cellDocListList.get(cellId).add(i);
		}
		return cellDocListList;
	}
	
	private List<List<Integer>> buildDocCellSTR(int sqrtN)
	{
		class DocLoc {
			int docId;
			Location loc;
			DocLoc(int docId, Location loc)
			{
				this.docId = docId;
				this.loc = loc;
			}
		}
		
		int numCell = sqrtN * sqrtN;
		
		List<DocLoc> docLocListX = new ArrayList<DocLoc>();
		for (int i = 0; i != this.numDoc; ++i)
			docLocListX.add(new DocLoc(i, this.docToLocArray[i]));
		
		Collections.sort(docLocListX, new Comparator<DocLoc>() {
			@Override
			public int compare(DocLoc a, DocLoc b) {
				//sort by x asc
				if (a.loc.getLocx() > b.loc.getLocx())
					return 1;
				else
					return -1;
			}
		});
		
		int cellId = 0;
		List<List<Integer>> cellDocListList = new ArrayList<List<Integer>>();
		for (int i = 0; i != numCell; ++i)
			cellDocListList.add(new ArrayList<Integer>());
		
		int num = 0;
		int numX = this.numDoc / sqrtN;
		int offsetX = 0;
		for (int xx = 0; xx < sqrtN; ++xx)
		{
			if (xx == sqrtN - 1)
				numX = this.numDoc - offsetX;
			List<DocLoc> docLocListY = new ArrayList<DocLoc>();
			for (int j = offsetX; j < offsetX + numX; ++j)
				docLocListY.add(docLocListX.get(j));
			offsetX += numX;
			
			Collections.sort(docLocListY, new Comparator<DocLoc>() {
				@Override
				public int compare(DocLoc a, DocLoc b) {
					//sort by y asc
					if (a.loc.getLocy() > b.loc.getLocy())
						return 1;
					else
						return -1;
				}
			});
			
			int numY = numX / sqrtN;
			int offsetY = 0;
			for (int yy = 0; yy < sqrtN; ++yy)
			{
				if (yy == sqrtN - 1)
					numY = numX - offsetY;
				for (int j = offsetY; j < offsetY + numY; ++j)
					cellDocListList.get(cellId).add(docLocListY.get(j).docId);
				num += cellDocListList.get(cellId).size();
				System.out.print(cellId + "(" + cellDocListList.get(cellId).size() + ") ");
				cellId++;
				offsetY += numY;
			}
			System.out.println();
		}
		assert(cellId == numCell);
		assert(num == this.numDoc);
		
		return cellDocListList;
	}
	
	private int [] buildDocClusterArray(int N, int M, File docClusterFile, String method) throws IOException
	{
		int numCell = N * N;
		List<List<Integer>> cellDocListList = null;
		if (method.startsWith("STR"))
			cellDocListList = this.buildDocCellSTR(N);
		else if (method.startsWith("Cell"))
			cellDocListList = this.buildDocCell(N, numCell);
		else if (method.startsWith("Rand"))
			cellDocListList = this.buildDocCellRand(numCell);
		else
			Tools.printError("Wrong Document Clustering Method");
		
		assert(numCell == cellDocListList.size());
		
		int [] docToClusterArray = new int[numDoc];
		if (M <= 1)
		{
			for (int cellId = 0; cellId != numCell; ++cellId)
			{
				List<Integer> docList = cellDocListList.get(cellId);
				for (int jj = 0; jj != docList.size(); ++jj)
				{
					int docId = docList.get(jj);
					docToClusterArray[docId] = cellId;
				}
			}
			return docToClusterArray;
		}
		
		for (int i = 0; i != numCell; ++i)
		{
			int clusterOffset = i * M;
			List<Integer> docList = cellDocListList.get(i);
			System.out.println("****Clustering Cell" + i + "/" + numCell + " (" + docList.size() + ")****");
			if (docList.isEmpty())
				continue;
			
			int [] clusterArray = null;
			if (method.endsWith("kMeans"))
			{
				int maxIter = N == 1 ? 10 : 100;
				if (M > 100)
					maxIter = 1;
				clusterArray = this.clusterCellkMeans(docList, M, maxIter);
			}
			else if (method.endsWith("kMedoid"))
				clusterArray = this.clusterCellPAM(docList, N, M, i);
			else if (method.endsWith("Rand"))
				clusterArray = this.clusterCellRand(docList, M);
			else
				Tools.printError("Wrong Query Clustering Method");
			
			for (int jj = 0; jj != docList.size(); ++jj)
			{
				int docId = docList.get(jj);
				int cluster = clusterArray[jj];
				docToClusterArray[docId] = clusterOffset + cluster;
			}
		}
		
		PrintWriter writer = new PrintWriter(docClusterFile);
		for (int i = 0; i != numDoc; ++i)
			writer.println(docToClusterArray[i]);
		writer.close();
		
		return docToClusterArray;
	}
	
	private int [] readClusterArray(int numData, File file) throws IOException
	{
		int [] dataToClusterArray = new int[numData];
		Scanner in = new Scanner(file);
		for (int i = 0; i != numData; ++i)
			dataToClusterArray[i] = in.nextInt();
		in.close();
		return dataToClusterArray;
	}
	
	private int [] clusterCellRand(List<Integer> docList, int M)
	{
		int numData = docList.size();
		int [] clusterArray = new int[numData];
		for (int i = 0; i != numData; ++i)
			clusterArray[i] = (int)(Math.random() * M);
		
		return clusterArray;
	}
	
	private int [] clusterCellPAM(List<Integer> docList, int N, int M, int cellId) throws IOException
	{
		SMatrix matrix = this.sDocClusterMatrix;
		
		int numDoc = docList.size();
		if (numDoc == 0)
			return null;
		
		Timer timer = new Timer();
		timer.start();
		
		Tools.mkdir(this.clusterPrefix + "matrix/");
		File docDocMatrixFile = new File(this.clusterPrefix + "matrix/doc_doc_N"
				+ N + "_M" + M + "_" + cellId + ".matrix");
		
		SMatrix docDocSim = null;
		if (docDocMatrixFile.isFile())
			docDocSim = new SMatrix(docDocMatrixFile, false);
		else
		{
			docDocSim = new SMatrix(numDoc, numDoc);
			for (int i = 0; i < numDoc; ++i)
			{
				if (i % 1000 == 0)
					System.out.println(i + "/" + numDoc);
				for (int j = i; j < numDoc; ++j)
				{
					SVector vecA = matrix.getRowVector(docList.get(i));
					SVector vecB = matrix.getRowVector(docList.get(j));
					double sim = vecA.cosineSimilarity(vecB);
					if (sim > 0)
						docDocSim.set(i, j, sim);
				}
			}
			System.out.println("##SimCost = " + timer.stop() + " seconds##");
			docDocSim.write(docDocMatrixFile, false);
		}

		timer.start();
		PAM pam = new PAM();
		int [] result = pam.cluster(docList, numDoc, M, docDocSim);
		System.out.println("##ClusterCost = " + timer.stop() + " seconds##");
		
		return result;
	}
	
	private int [] clusterCellkMeans (List<Integer> docList, int M, int maxIter)
	{
		if (docList.size() == 0)
			return null;
		
		SMatrix matrix = this.sDocClusterMatrix;
		return this.kMeans(matrix, docList, M, maxIter);
	}
	
//	private double cosineSimilarity (SimpleSparseVector vec1, SimpleSparseVector vec2)
//	{
//		double product = 0, norm1 = 0, norm2 = 0;
//		SimpleSparseVector smallerVec, largerVec;
//		if (vec1.numEntries() < vec2.numEntries())
//		{
//			smallerVec = vec1;
//			largerVec = vec2;
//		}
//		else
//		{
//			smallerVec = vec2;
//			largerVec = vec1;
//		}
//		
//		for (Entry<Integer, Double> vecEntry : smallerVec)
//		{
//			int index = vecEntry.getKey();
//			double val1 = vecEntry.getValue();
//			double val2 = largerVec.get(index);
//			product += val1 * val2;
//		}
//		
//		norm1 = vec1.normL2();
//		norm2 = vec2.normL2();
//		
//		double sim = product/norm1/norm2;
//		assert(sim < 1.1);	//there might be small error due to precision
//		if (sim > 1)
//			sim = 1;
//		
//		return sim;
//	}
	
	private int [] kMeans (SMatrix matrix, List<Integer> docList, int K, int maxIter)
	{
		int num = docList.size();
		int [] clusterArray = new int[num];
		
		if (num < 100)	//in case of less than 100 data, put all into the 0th cluster
			return clusterArray;
		
		SVector [] center = new SVector[K];
//		for (int kk = 0; kk != K; ++kk)
//		{
//			center[kk] = new SVector();
//			for (int i = 0; i != matrix.numColumns(); ++i)
//				center[kk].set(i, Math.random());
//		}
		Set<Integer> randSet = new HashSet<Integer>();
		for (int kk = 0; kk != K; ++kk)
		{
			int randId = -1;
			while (true)
			{
				randId = (int) (Math.random() * num);
				if (randSet.contains(randId) == false)
				{
					randSet.add(randId);
					break;
				}
			}
			center[kk] = matrix.getRowVector(randId);
		}
		
		int iter = 0;
		//int capacity = num / K + 1;
		
		while (true)
		{			
			if (iter >= maxIter)
				break;
			
			System.out.print("iter" + iter + ": ");

			int [] clusterNum = new int[K];
			for (int i = 0; i != num; ++i)
			{
				SVector vec = matrix.getRowVector(i);
				clusterArray[i] = -1;
				double maxSim = -1;
				for (int kk = 0; kk != K; ++kk)
				{
//					if (clusterNum[kk] >= capacity)
//						continue;
					double sim = vec.cosineSimilarity(center[kk]);
					if (sim > maxSim)
					{
						maxSim = sim;
						if (clusterArray[i] != -1)
							clusterNum[clusterArray[i]]--;
						clusterArray[i] = kk;
						clusterNum[clusterArray[i]]++;
					}
				}
				
				if (maxSim == 0)
					clusterArray[i] = (int) (Math.random() * K);
			}
			
			SVector [] clusterVec = new SVector[K];
			for (int kk = 0; kk != K; ++kk)
				clusterVec[kk] = new SVector();
			
			for (int i = 0; i != num; ++i)
			{
				//clusterNum[clusterArray[i]]++;
				clusterVec[clusterArray[i]].add(matrix.getRowVector(i));
			}
			
			boolean equal = true;
			for (int kk = 0; kk != K; ++kk)
			{
				clusterVec[kk].scale(1.0 / clusterNum[kk]);
				if (clusterVec[kk].equals(center[kk]) == false)
				{
					center[kk] = clusterVec[kk];
					equal = false;
				}
			}
			
			for (int kk = 0; kk != K; ++kk)
				System.out.print(clusterNum[kk] + " ");
			System.out.println();

			if (equal)
				break;
			
			iter++;
		}
		
		int [] clusterNum = new int[K];
		for (int i = 0; i != num; ++i)
			clusterNum[clusterArray[i]]++;
		
//		System.out.print("iter" + iter + ": ");
//		for (int kk = 0; kk != K; ++kk)
//			System.out.print(clusterNum[kk] + " ");
//		System.out.println();
		
		return clusterArray;
	}
	
	private int [] sBuildQueryClusterArray_Cell (int numQuery, 
			int numCluster, int [] docClusterArray, File queryClusterFile) throws IOException
	{
		//SMatrix matrix = this.sQueryClusterMatrix;
		SMatrix matrix = this.sRawQueryClusterMatrix;
		
		int [] queryToClusterArray = new int[numQuery];

		PrintWriter writer = new PrintWriter(queryClusterFile);
			
		for (int qId = 0; qId != numQuery; ++qId)
		{
			double [] clusterScore = new double[numCluster];
			SVector vec = matrix.getRowVector(qId);
			for (Entry<Integer, Double> vecEntry : vec)
			{
				int dId = vecEntry.getKey();
				int cId = docClusterArray[dId];
				clusterScore[cId] += 1;
			}
			
			int cluster = 0;
			double maxScore = clusterScore[0];
			for (int cId = 1; cId < numCluster; ++cId)
			{
				if (clusterScore[cId] > maxScore)
				{
					maxScore = clusterScore[cId];
					cluster = cId;
				}
			}
			queryToClusterArray[qId] = cluster;
			//System.out.println("query" + qId + ": " + cluster + "\t" + clusterScore[cluster]);
			writer.println(cluster);
		}

		writer.close();
		
		return queryToClusterArray;
	}
	
	public int [] getDocCellArray (int numDoc, int sqrtN)
	{
		int [] docToCellArray = new int[numDoc];
		for (int i = 0; i != numDoc; ++i)
			docToCellArray[i] = Tools.locToCellID(this.docToLocArray[i], sqrtN, 1.0/sqrtN);
		return docToCellArray;
	}
	
	public int [] getQueryClusterArray_Rand (int numQuery, int numCluster)
	{
		int [] queryToClusterArray = new int[numQuery];
		int i = 0;
		int cluster = 0;
		while (i < numQuery)
		{
			if (cluster == numCluster)
				cluster = 0;
			queryToClusterArray[i++] = cluster++;
		}
		return queryToClusterArray;
	}
	
	private int [] getRandomCluster (int numData, int numCluster)
	{
		int [] cluster = new int[numData];
		for (int i = 0; i < numData; ++i)
			cluster[i] = (int) (Math.random() * numCluster);
		
		return cluster;
	}
}
