package sugg.cluster;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import sugg.matrix.SMatrix;

/**
 * Partitioning Around Medoids. Cluster data elements by aiming to minimize the average
 * dissimilarity of objects to their closest selected element (medoid).
 * Data elements are indexable.
 * Independent of Cytoscape.
 * NB   This class is an implementation of the PAM algorithm described in the course notes
 *      at <www.cs.umb.edu/cs738/pam1.pdf> (accessed 2012-07-31).
 *      There appears to be different variants of the PAM algorithm varying in details
 *      within the build and swap phases.
 *      The original algorithm is described in chapter 2 of Kaufman and Rousseeuw (1990),
 *      whose original Fortran code was translated and augmented as part of the cluster R package.
 *      The cluster results from current implementation can differ from the implementation 
 *      in R's cluster::pam. (See PAMTest for details.)
 * @author djh.shih
 * @comment
 *
 * shuyao: modified for specialized usage
 */
public class PAM {
	private int numData;
	private int numCluster;
	private SMatrix sim;
	//private SMatrix matrix;
	
	// ordered index of element subset
	// all immediate data and output should have the same size as idx
	// (only the input BaseMatrix data has the full original set of data elements)
	// (indexing of DistanceMatrix is handled by the class itself)
	int[] idx;
	
	// distance between element and closest medoid
	protected double[] nearestDistances;
	// distance between element and second closest medoid
	double[] nextNearestDistances;
	
	// nearest medoid of each element
	int[] nearestMedoids;
	// next-nearest medoid of each element 
	int[] nextNearestMedoids;
	
	// set of medoids
	HashSet<Integer> medoids;
	
	// set of non-meoids (maintain for finding swap candidates)
	HashSet<Integer> nonmedoids;
	
	// set of all indexed elements
	// required since Java's HashSet cannot use native types
	Integer[] elements;
	
	int maxSwaps = 100;

	public int [] cluster(List<Integer> data, int numData, int numCluster, SMatrix matrix)
	//public int [] cluster(List<Integer> data, int numData, int numCluster, float [][] sim)
	{	
		int [] clusterArray = new int[numData];

		if (numData < 100)	//in case of less than 100 data, put all into the 0th cluster
			return clusterArray;
		
		//this.simMatrix = simMatrix;
		this.numData = numData;
		this.numCluster = numCluster;
		
		//this.matrix = matrix;
		//this.sim = new SMatrix(numData, numData);
		this.sim = matrix;
		
		assert(data.size() == numData);
		
		this.idx = new int[numData];
		for (int i = 0; i != numData; ++i)
			this.idx[i] = data.get(i);
		
		initialize();
		//buildPhraseRand();
		buildPhrase();
		//swapPhase();
		
		Map<Integer, Integer> medoidClusterMap = new HashMap<Integer, Integer>();
		int [] clusterCountArray = new int[numCluster];
		int kk = 0;
		for (Integer medoid : this.medoids)
			medoidClusterMap.put(medoid, kk++);
		
		for (int i = 0; i != this.numData; ++i)
		{
			int cluster = medoidClusterMap.get(this.nearestMedoids[i]);
			clusterCountArray[cluster]++;
			clusterArray[i] = cluster;
		}
		
		for (int i = 0; i != this.numCluster; ++i)
			System.out.print(clusterCountArray[i] + " ");
		System.out.println();
		
		return clusterArray;
	}
	
//	/**
//	 * Calculate the clustering cost: sum of distances to cluster medoids.
//	 * @return cost
//	 */
//	private double getCost() {
//		double c = 0;
//		for (int i = 0; i < nearestDistances.length; ++i) {
//			c += nearestDistances[i];
//		}
//		return c;
//	}
	
	private void initialize() {
		nearestDistances = new double[numData];
		nextNearestDistances = new double[numData];
		nearestMedoids = new int[numData];
		nextNearestMedoids = new int[numData];		
		elements = new Integer[numData];
		medoids = new HashSet<Integer>();
		nonmedoids = new HashSet<Integer>();
		
		for (int ii = 0; ii < numData; ++ii)
		{
			// initialize distances to infinity
			nearestDistances[ii] = nextNearestDistances[ii] = Double.POSITIVE_INFINITY;
			// initialize medoids to non-valid indices, s.t. unexpected bugs trigger indexing error
			nearestMedoids[ii] = nextNearestMedoids[ii] = -1;
			
			elements[ii] = new Integer(ii);
			
			// all (indexed) data elements are initially non-medoids
			nonmedoids.add( elements[ii] );
		}
	}
	
	private double distance (int i, int j)
	{
		double similarity = this.similarity(i, j);
		assert(similarity <= 1);
		return 1.0 - similarity;
	}
	
	private double similarity (int i, int j)
	{
		if (i > j)
		{
			int temp = i;
			i = j;
			j = temp;
		}
		
		return this.sim.get(i, j);
		
//		double sim = this.sim.get(i, j);
//		if (sim > 0)
//			return sim;
//		else
//		{
//			sim = this.cosineSimilarity(
//					matrix.getRowVector(this.idx[i]), 
//					matrix.getRowVector(this.idx[j]));
//			if (sim > Tools.EP)
//				this.sim.set(i, j, sim);
//			return sim;
//		}
	}
	
	//shuyao: build intitial medoids randomly
	@SuppressWarnings("unused")
	private void buildPhraseRand()
	{
		Set<Integer> randSet = new HashSet<Integer>();
		for (int i = 0; i != numCluster; ++i)
		{
			int rand = -1;
			while (true)
			{
				rand = (int) (Math.random() * numData);
				if (randSet.contains(rand) == false)
				{
					randSet.add(rand);
					addMedoid(rand);
					break;
				}
			}
		}
		
		if (numCluster != medoids.size())
		{
			throw new RuntimeException("Expected error in BUILD phase: Number of medoids does not match parameter k.");
		}
	}
	
	/**
	 * BUILD phase. Select a initial set of k medoids.
	 */
	private void buildPhrase() {
		this.maxSwaps = 1;		//TODO
		
		// select first medoid
		
		// find element with minimum total distance to all other elements
		double[] totalDistances = new double[numData];
		for (int ii = 0; ii < numData; ++ii) {
			// sum distances to all other elements
			// assume distance to itself is 0
			double d = 0;
			for (int jj = 0; jj < numData; ++jj) {
				d += distance(ii, jj);
			}
			totalDistances[ii] = d;
		}
		double minDistance = totalDistances[0];
		int minIndex = 0;
		for (int ii = 0; ii < numData; ++ii) {
			if (totalDistances[ii] < minDistance) {
				minDistance = totalDistances[ii];
				minIndex = ii;
			}
		}
		// add element to medoid set
		addMedoid(minIndex);
		
		
		// select remaining k - 1 medoids		
		double[] gains = new double[numData];
		
		for (int kk = 1; kk < numCluster; ++kk) {		
			// consider each i as medoid candidate
			for (int ii = 0; ii < numData; ++ii) {
				// if ii is already a medoid, it has negative gain to prevent it from being selected again
				if (medoids.contains(elements[ii])) {
					gains[ii] = -1.0;
				} else {
					double gain = 0;
					// for each non-medoid j != i, calculate the gain
					for (int jj = 0; jj < numData; ++jj) {
						if (jj == ii || medoids.contains(elements[jj]) ) continue;
						if (nearestDistances[jj] > distance(ii, jj)) {
							// add i will improve j's nearest distances
							// (if selected, i will be the new nearest neighbour of j)
							gain += nearestDistances[jj] - distance(ii, jj);
						}
					}
					gains[ii] = gain;
				}
			}
			// select candidate with maximum gain
			double maxGain = Double.NEGATIVE_INFINITY;
			int maxIndex = -1;
			for (int ii = 0; ii < numData; ++ii) {
				if (gains[ii] > maxGain) {
					maxGain = gains[ii];
					maxIndex = ii;
				}
			}
			// add element to medoid set
			addMedoid(maxIndex);
			
		}
		
		// check that the number of medoids match the expected
		if (numCluster != medoids.size())
		{
			throw new RuntimeException("Expected error in BUILD phase: Number of medoids does not match parameter k.");
		}
		
	}
	
	/**
	 * SWAP phase. Attempt to improve clustering quality by exchanging medoids with non-medoids.
	 */
	@SuppressWarnings("unused")
	private void swapPhase() {
		boolean notConverged = true;
		boolean continueLoop = true;
		int nSwaps = 0;
		
		while (notConverged && continueLoop) {
			notConverged = false;
			//continueLoop = false;
	
			Iterator<Integer> medoidIter = medoids.iterator();
			while (medoidIter.hasNext() && continueLoop)
			{
				//for each medoid i
				int ii = medoidIter.next().intValue();
				
				Iterator<Integer> nonMedoidIter = nonmedoids.iterator();
				while (nonMedoidIter.hasNext())
				{
					//for each non-medoid h
					int hh = nonMedoidIter.next().intValue();
				
					// Consider swapping medoid i and nonmedoid h
					// by calculating gains by all other elements
					
					// Calculate cumulative change to distance to nearest medoid for all nonmedoids j != h
					double change = 0;
					Iterator<Integer> nonMedoidIter2 = nonmedoids.iterator();
					while (nonMedoidIter2.hasNext())
					{
						int jj = nonMedoidIter2.next().intValue();
						if (jj == hh) 
							continue;
					
						double d = nearestDistances[jj];
						if (distance(ii, jj) > d) {
							// if removed, i will have no impact
							if (distance(jj, hh) < d) {
								// if selected, h will improve nearest distance for j
								change += distance(jj, hh) - d;
							}
						} else {
							// i cannot be closer than the nearest neighbour for j;
							// therefore, distances[i][j] == d
							// and i is currently the nearest neighbour for j
							double e = nextNearestDistances[jj];
							if (distance(jj, hh) < e) {
								// if i and h are swapped, h will become the nearest neighbour
								// nearest distance for j may improve or worsen
								change += distance(jj, hh) - d;
							} else {
								// if i is removed, the current next-nearest of j will be promoted to nearest
								change += e - d;
							}
						}
					}
					
					if (change < -2) {	//shuyao: changed from 0 to -2
						// distance to nearest medoid summed over all nonmedoids is improved: swap
						swap(hh, ii);
						System.out.print("Swap " + hh + " and " + ii + " for change = " + change + "\n");
						
						// non-convergence if any swap occurs, up to a maximum number of swaps (to guard against swap cycles)
						if (nSwaps++ < maxSwaps) {
							notConverged = true;
						} else {
							continueLoop = false;
						}
						
						// reset iterator
						medoidIter = medoids.iterator();
						// break out of inner loop to consider next medoid
						break;
					}
				}
				
			}
			
		}
	}
	
	private void addMedoid(int add) {
		medoids.add( elements[add] );
		nonmedoids.remove( elements[add] );
		updateNearest(add, -1);
	}
	
	private void swap(int add, int remove) {
		medoids.add( elements[add] );
		nonmedoids.remove( elements[add] );
		medoids.remove( elements[remove] );
		nonmedoids.add( elements[remove] );
		updateNearest(add, remove);
	}
	
	/**
	 * Update nearest and next-nearest distances.
	 * Does not check whether {@code added} or {@ removed} have been added to or removed from the medoid set.
	 * @param added Index of element added to medoid set (-1 for none)
	 * @param removed Index of element removed from medoid set (-1 for none)
	 */
	private void updateNearest(int added, int removed) {
		if (added >= 0) {
			// added index is valid
			
			// check if any nearest distance improves
			for (int ii = 0; ii < numData; ++ii) {
				double d = distance(ii, added);
				if (d < nearestDistances[ii]) {
					// element i is nearer to added medoid than previous nearest: update
					double oldDistance = nearestDistances[ii];
					int oldMedoid = nearestMedoids[ii];
					nearestMedoids[ii] = added;
					nearestDistances[ii] = d;
					// pump nearest distance to next-nearest distance
					nextNearestMedoids[ii] = oldMedoid;
					nextNearestDistances[ii] = oldDistance;
				} else if (d < nextNearestDistances[ii]) {
					// element i is nearer to added medoid than previous next-nearest: update
					nextNearestMedoids[ii] = added;
					nextNearestDistances[ii] = d;
				}
			}
			
		}
		
		if (removed >= 0) {
			// removed index is valid
			
			// check if the removed medoid is the nearest or next-nearest of any element
			for (int ii = 0; ii < numData; ++ii) {
				if (nearestMedoids[ii] == removed) {
					// promote next-nearest to nearest
					nearestMedoids[ii] = nextNearestMedoids[ii];
					nearestDistances[ii] = nextNearestDistances[ii];
					// find new next-nearest
					updateNextNearest(ii);
				} else if (nextNearestMedoids[ii] == removed) {
					// find new next-nearest
					updateNextNearest(ii);
				}
			}
			
		}
	}
	
	/**
	 * Update next nearest for element i.
	 * Assume nearest medoid is already set.
	 * @param ii element index to be updated
	 */
	private void updateNextNearest(int ii) {
		int nearestMedoid = nearestMedoids[ii];
		
		// find the next-nearest
		Iterator<Integer> it = medoids.iterator();
		double minDistance = Double.POSITIVE_INFINITY;
		int nextNearestMedoid = -1;
		while (it.hasNext()) {
			int jj = it.next().intValue();
			// ignore if j is the nearestMedoid, since we are interested in the next-nearest
			if (jj == nearestMedoid) continue;
			if (distance(ii, jj) < minDistance) {
				minDistance = distance(ii, jj);
				nextNearestMedoid = jj;
			}
		}
	
		// update
		nextNearestDistances[ii] = minDistance;
		nextNearestMedoids[ii] = nextNearestMedoid;
	}

}
