package sugg.matrix;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;

import sugg.tools.Timer;
import sugg.tools.Tools;

import java.util.Scanner;

public class SMatrix {
	private int numRow;
	private int numCol;
	private long numEntry;
	
	private Map<Integer, SVector> data;
	
	public SMatrix (int numRow, int numCol) 
	{
		this.numRow = numRow;
		this.numCol = numCol;
		this.numEntry = 0;
		this.data = new HashMap<Integer, SVector>();
	}
	
	public SMatrix (File matrixFile, boolean transposed) throws IOException
	{
		Timer timer = new Timer();
		timer.start();
		Scanner in = new Scanner(matrixFile);
		in.nextLine();
		if (transposed)
		{
			this.numCol = in.nextInt();
			this.numRow = in.nextInt();
		}
		else
		{
			this.numRow = in.nextInt();
			this.numCol = in.nextInt();
		}
		long numEntry = in.nextLong();
		this.data = new HashMap<Integer, SVector>();
		
		System.out.println("Reading Matrix: numEntry = " + numEntry);
		
		for (long i = 0; i < numEntry; ++i)
		{
			int row = in.nextInt();
			int col = in.nextInt();
			double val = in.nextDouble();
			if (transposed)
				this.set(col, row, val);
			else
				this.set(row, col, val);
		}
		
		assert(this.numEntry == numEntry || this.numEntry == numEntry - 1);
		//the [0][0] entry might be 0 and thus omitted
		
		in.close();
		System.out.println("##Matrix read in " + timer.stop() + " seconds##");
	}
	
	public void setRowVector (int row, SVector vec)
	{
		this.data.put(row, vec);
	}
	
	public SVector getRowVector (int row)
	{
		if (this.data.containsKey(row) == false)
			this.data.put(row, new SVector());
		return this.data.get(row);
	}
	
	public void write (File matrixFile, boolean transposed) throws IOException
	{
		PrintWriter writer = new PrintWriter(matrixFile);
		writer.println("%%MatrixMarket matrix coordinate real general");
		if (transposed)
			writer.format(Locale.ENGLISH, "%10d %10d %19d\n", this.numCol, this.numRow, this.numEntry);
		else
			writer.format(Locale.ENGLISH, "%10d %10d %19d\n", this.numRow, this.numCol, this.numEntry);
		for (int row = 0; row < this.numRow; ++row)
		{
			if (this.data.containsKey(row) == false)
				continue;
			SVector rowVector = this.data.get(row);
			for (Entry<Integer, Double> valEntry : rowVector)
			{
				int col = valEntry.getKey();
				double val = valEntry.getValue();
				if (transposed)
					writer.format(Locale.ENGLISH, "%10d %10d % .12e\n", col, row, val);
				else
					writer.format(Locale.ENGLISH, "%10d %10d % .12e\n", row, col, val);
			}
		}
		writer.close();
	}
	
	public boolean contains (int row, int col)
	{
		if (this.data.containsKey(row) == false)
			return false;
		return this.data.get(row).contains(col);
	}
	
	public double get (int row, int col)
	{
		if (this.data.containsKey(row) == false)
			return 0.0;
			//this.data.put(row, new SVector());
		return this.data.get(row).get(col);
	}
	
	public void set (int row, int col, double val)
	{
		if (val == 0)
			return;
		if (this.data.containsKey(row) == false)
			this.data.put(row, new SVector());
		if (this.data.get(row).contains(col) == false)
			this.numEntry++;
		this.data.get(row).set(col, val);
	}
	
	public void add (int row, int col, double val)
	{
		this.set(row, col, val + this.get(row, col));
	}
	
	public SMatrix mult (SMatrix other)
	{
		if (this.numRow != other.numCol || this.numCol != other.numRow)
			Tools.printError("Wrong matrix multiplication input");
		SMatrix otherT = other.transpose();
		SMatrix result = new SMatrix(this.numRow, this.numRow);
		
		for (int i = 0; i < this.numRow; ++i)
		{
			SVector rowVec = this.getRowVector(i);
			for (int j = 0; j < other.numCol; ++j)
			{
				SVector colVec = otherT.getRowVector(j);
				double sum = 0;
				for (Entry<Integer, Double> rowEntry : rowVec)
				{
					int col = rowEntry.getKey();
					double val = rowEntry.getValue() * colVec.get(col);
					sum += val;
				}
				result.set(i, j, sum);
			}
		}
		
		return result;
	}
	
	public int numRows()
	{
		return this.numRow;
	}
	
	public int numColumns()
	{
		return this.numCol;
	}
	
	public long numEntries()
	{
		return this.numEntry;
	}
	
	public void normalizeRow(int row)
	{
		this.getRowVector(row).normalize();
	}
	
	public SMatrix transpose()
	{
		SMatrix transMatrix = new SMatrix(this.numCol, this.numRow);
		for (int row = 0; row != this.numRow; ++row)
		{
			SVector rowVector = this.getRowVector(row);
			for (Entry<Integer, Double> entry : rowVector)
			{
				int col = entry.getKey();
				double val = entry.getValue();
				transMatrix.set(col, row, val);
			}
		}
		return transMatrix;
	}
	
	public SMatrix getCosSimMatrix(String filePath) throws IOException
	{
		File file = new File(filePath);
		SMatrix matrix = null;
		if (file.isFile())
		{
			System.out.println("Reading Cosine Similarity Matrix (" + filePath + ")");
			matrix = new SMatrix(file, false);
		}
		else
		{
			System.out.println("Calculating Cosine Similarity Matrix (" + filePath + ")");
			Timer timer = new Timer();
			timer.start();
			double cost = 0;
			matrix = new SMatrix(this.numRow, this.numRow);
			for (int i = 0; i != this.numRow; ++i)
			{
				if (i % 100 == 0)
				{
					cost += timer.stop();
					Tools.printProgress(i, numRow, cost);
					timer.start();
				}
				for (int j = i + 1; j != this.numRow; ++j)
				{
					double sim = this.getRowVector(i).cosineSimilarity(this.getRowVector(j));
					matrix.set(i, j, sim);
				}
			}
			matrix.write(file, false);
		}
		
		return matrix;
	}
	
	public void exportAsCSV (String outputPath) throws IOException
	{
		assert(this.numRow == this.numCol);
		PrintWriter writer = new PrintWriter(new File(outputPath));
		for (int i = 0; i != this.numRow; ++i)
			writer.print(";" + i);
		writer.println();
		for (int i = 0; i != this.numRow; ++i)
		{
			if (i % 1000 == 0)
				System.out.println(i);
			writer.print(i);
			for (int j = 0; j < i; ++j)
				writer.print(";" + this.get(j, i));
			for (int j = i; j < this.numRow; ++j)
				writer.print(";" + this.get(i, j));
			writer.println();
		}
		writer.close();
	}
	
	public void print()
	{
		System.out.println("NumRow = " + this.numRow + " NumCol = " + this.numCol);
		for (int i = 0; i < this.numRow; ++i)
		{
			SVector vec = this.getRowVector(i);
			for (int j = 0; j < this.numCol; ++j)
				System.out.printf("%.2f\t", vec.get(j));
			System.out.println();
		}
	}
}
