package sugg.matrix;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import sugg.tools.Tools;

import java.util.Set;

public class SVector implements Iterable<Entry<Integer, Double>> {
	private Map<Integer, Double> vector;
	private double normL2 = -1;
	private double normL1 = -1;
	private double normInfty = -1;
	
	public SVector()
	{
		this.vector = new HashMap<Integer, Double>();
	}
	
	public int size()
	{
		return this.vector.size();
	}
	
	public double get(int index)
	{
		if (this.vector.containsKey(index) == false)
			return 0;
		else
			return this.vector.get(index);
	}
	
	public boolean contains(int index)
	{
		return this.vector.containsKey(index);
	}
	
	public void set (int index, double value)
	{
		if (value == 0.0)
			return;
		this.vector.put(index, value);
	}
	
	public void add (int index, double value)
	{
		this.set(index, this.get(index) + value);
	}
	
	public double normL2()
	{
		if (this.normL2 < 0)
		{
			double norm2 = 0;
			for (Entry<Integer, Double> vecEntry : this.vector.entrySet())
			{
				double val = vecEntry.getValue();
				norm2 += val * val;
			}
			this.normL2 = Math.sqrt(norm2);
		}
		return this.normL2;
	}
	
	public double normL1()
	{
		if (this.normL1  < 0)
		{
			double norm = 0;
			for (Entry<Integer, Double> vecEntry : this.vector.entrySet())
				norm += vecEntry.getValue();
			this.normL1 = norm;
			assert(this.normL1 >= 0);
		}
		return this.normL1;
	}
	
	public double normInfty()
	{
		if (this.normInfty < 0)
		{
			for (Entry<Integer, Double> vecEntry : this)
			{
				if (vecEntry.getValue() > this.normInfty)
					this.normInfty = vecEntry.getValue();
			}
		}
		return this.normInfty;
	}
	
	public double getAverage()
	{
		return this.normL1() / this.vector.size();
	}
	
	public void add (SVector otherVec)
	{
		for (Entry<Integer, Double> vecEntry : otherVec.vector.entrySet())
		{
			int index = vecEntry.getKey();
			double val = vecEntry.getValue();
			this.set(index, val + this.get(index));
		}
	}
	
	public void scale (double mul)
	{
		for (Entry<Integer, Double> vecEntry : this.vector.entrySet())
		{
			int index = vecEntry.getKey();
			double val = vecEntry.getValue();
			val *= mul;
			this.set(index, val);
		}
	}
	
	public boolean equals (SVector otherVec)
	{
		for (Entry<Integer, Double> vecEntry : otherVec.vector.entrySet())
		{
			int index = vecEntry.getKey();
			double val = vecEntry.getValue();
			if (this.get(index) != val)
				return false;
		}
		return true;
	}
	
	public void normalize01()
	{
	    double maxScore = -Double.MAX_VALUE;
	    double minScore = Double.MAX_VALUE;
		for (Entry<Integer, Double> vecEntry : this.vector.entrySet())
		{
			double val = vecEntry.getValue();
    		if (val > maxScore)
    			maxScore = val;
    		if (val < minScore)
    			minScore = val;
		}
		
		if (maxScore > minScore)
		{
			for (Entry<Integer, Double> vecEntry : this.vector.entrySet())
			{
				int index = vecEntry.getKey();
				double val = vecEntry.getValue();
				this.vector.put(index, (val - minScore) / (maxScore - minScore));
			}
		}
	}
	
	public void normalize1()
	{
		double maxScore = this.normInfty;
		this.scale(1.0/maxScore);
	}
	
	public void normalize()
	{
		double norm = 0;
		for (Entry<Integer, Double> vecEntry : this.vector.entrySet())
		{
			double val = vecEntry.getValue();
			norm += val;
		}

		for (Entry<Integer, Double> vecEntry : this.vector.entrySet())
		{
			int index = vecEntry.getKey();
			double val = vecEntry.getValue();
			this.vector.put(index, val / norm);
		}
	}
	
	public int numEntries()
	{
		return this.vector.size();
	}
	
	public double squareSimilarity (SVector other)
	{
		double sum = 0;
		Set<Integer> visited = new HashSet<Integer>();
		for (Entry<Integer, Double> vecEntry : this)
		{
			int key = vecEntry.getKey();
			double thisVal = vecEntry.getValue();
			double otherVal = other.get(key);
			if (otherVal > 0)
				visited.add(key);
			
			sum += Math.pow(thisVal - otherVal, 2);
			//System.out.print("(" + thisVal + ", " + otherVal + ") ");
		}
		//System.out.println();
		for (Entry<Integer, Double> vecEntry : other)
		{
			int key = vecEntry.getKey();
			double otherVal = vecEntry.getValue();
			if (visited.contains(key) == false)
				sum += otherVal * otherVal;
		}
		
		if (sum == 0)
		{
			Tools.printError("sum = 0");
			return Double.MAX_VALUE;
		}

		return 1.0/sum;
	}
	
	public double cosineSimilarity (SVector other)
	{
		double product = 0, norm1 = 0, norm2 = 0;
		SVector smallerVec, largerVec;
		if (this.numEntries() < other.numEntries())
		{
			smallerVec = this;
			largerVec = other;
		}
		else
		{
			smallerVec = other;
			largerVec = this;
		}
		
		for (Entry<Integer, Double> vecEntry : smallerVec)
		{
			int index = vecEntry.getKey();
			double val1 = vecEntry.getValue();
			double val2 = largerVec.get(index);
			product += val1 * val2;
		}
		
		if (product == 0)
			return 0;
		
		norm1 = smallerVec.normL2();
		norm2 = largerVec.normL2();
		
		double sim = product/norm1/norm2;
		assert(sim < 1.1);	//there might be small error due to precision
		if (sim > 1)
			sim = 1;
		
		return sim;
	}

	@Override
	public Iterator<Entry<Integer, Double>> iterator() {
		return this.vector.entrySet().iterator();
	}
	
	public void clear()
	{
		this.vector.clear();
		this.normL2 = -1;
	}
}
