package sugg.tools;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class SortedMap <K, V> {
	
	public class KeyValuePair implements Comparable<KeyValuePair> {
		public K key;
		public V value;
		
		public KeyValuePair(K key, V value)
		{
			this.key = key;
			this.value = value;
		}

		@Override
		public int compareTo(KeyValuePair o) {
			if (comparator.compare(this.value, o.value) < 0)
				return 1;
			else if (comparator.compare(this.value, o.value) > 0)
				return -1;
			else
				return 0;
		}
	}
	
	private Map<K, KeyValuePair> map;
	private List<KeyValuePair> list;
	private Comparator<V> comparator;
	
	private Map<K, KeyValuePair> lazyMap;
	//private Map<K, KeyValuePair> trashMap;
	
	private double sortCost;
	private double putCost;
	private double removeCost;
	private Timer timer;
	
	public SortedMap (Comparator<V> comparator)
	{
		this.timer = new Timer();
		this.sortCost = 0;
		this.comparator = comparator;
		this.map = new HashMap<K, KeyValuePair>();
		this.list = new ArrayList<KeyValuePair>();
		this.lazyMap = new HashMap<K, KeyValuePair>();
		//this.trashMap = new HashMap<K, KeyValuePair>();
	}
	
	public K getKey(int rank)
	{
		int index = this.size() - 1 - rank;
		return this.list.get(index).key;
	}
	
	public V getValue(int rank)
	{
		int index = this.size() - 1 - rank;
		return this.list.get(index).value;
	}
	
	public int size()
	{
		return this.list.size();
	}
	
	public boolean empty()
	{
		return this.list.size() == 0;
	}
	
	public K topKey()
	{
		return this.list.get(this.list.size() - 1).key;
	}
	
	public V topValue()
	{
		return this.list.get(this.list.size() - 1).value;
	}
	
	public void pop()
	{
		this.map.remove(this.topKey());
		this.list.remove(this.list.size() - 1);	//last is largest
	}
	
	public void put(K key, V value, boolean main)
	{
		if (main == true)	//put into main
		{
			this.remove(this.lazyMap, null, key);
			this.put(this.map, this.list, key, value);
		}
		else	//put into lazy list, won't be sorted
		{
			this.remove(this.map, this.list, key);
			this.put(this.lazyMap, null, key, value);
		}
	}
	
	private void put (Map<K, KeyValuePair> map, List<KeyValuePair> list, K key, V value)
	{
		this.timer.start();
		if (map.containsKey(key))
		{
			KeyValuePair keyvalue = map.get(key);
			keyvalue.value = value;
			if (keyvalue.key.equals(key) == false)
			{
				Tools.printError("Key value inconsistent. key = " + key + " value = " + value + 
						" oldkey = " + keyvalue.key);
				System.exit(1);
			}
		}
		else
		{
			KeyValuePair keyvalue = new KeyValuePair(key, value);
			if (list != null)
				list.add(keyvalue);
			map.put(key, keyvalue);
		}
		this.putCost += this.timer.stop();
	}
	
	private KeyValuePair remove (Map<K, KeyValuePair> map, List<KeyValuePair> list, K key)
	{
		if (map.containsKey(key))
		{
			this.timer.start();
			KeyValuePair keyvalue = map.get(key);
			map.remove(key);
			if (list != null)
				list.remove(keyvalue);
			this.removeCost += this.timer.stop();
			return keyvalue;
		}
		return null;
	}
	
	public V get(K key)
	{
		if (this.map.containsKey(key) == true)
			return this.map.get(key).value;
		if (this.lazyMap.containsKey(key) == true)
			return this.lazyMap.get(key).value;
		return null;
	}
	
	public void remove(K key)
	{
		KeyValuePair keyValue = this.remove(this.map, this.list, key);
		if (keyValue == null)
			keyValue = this.remove(this.lazyMap, null, key);
	}
	
	public void sort()
	{
		this.timer.start();
		System.setProperty("java.util.Arrays.useLegacyMergeSort", "true");
		Collections.sort(this.list);
		this.sortCost += this.timer.stop();
	}
		
	public double getSortCost()
	{
		return this.sortCost;
	}
	
	public double getPutCost()
	{
		return this.putCost;
	}
	
	public double getRemoveCost()
	{
		return this.removeCost;
	}
	
	public void print(double threshold)
	{
		for (int i = (this.list.size() - 1); i >= 0; i--)
		{
			KeyValuePair keyvalue = this.list.get(i);
			if ((Double)(keyvalue.value) < threshold)
				break;
			System.out.printf("%.3f(%d) ", keyvalue.value, keyvalue.key);
		}
		System.out.println();
	}
}
