package sugg.tools;

import java.io.File;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.tartarus.snowball.ext.PorterStemmer;

import edu.mit.jwi.IDictionary;
import edu.mit.jwi.item.IIndexWord;
import edu.mit.jwi.item.POS;
import edu.mit.jwi.morph.IStemmer;
import sugg.matrix.SVector;
import sugg.objects.Location;

public class Tools
{
	public static final double SQRT2 = 1.414;
	public static final double EP = 0.0000001;
	
	public static boolean equal (double x, double y)
	{
		return (Math.abs(x - y) < EP);
	}
	
	public static void printArray (double[] docClusterWeight)
	{
		for (double value : docClusterWeight)
			System.out.print(value + " ");
		System.out.println();
	}
	
	public static double averagePrecision (List<Integer> resultList, List<Integer> truthList)
	{
		double ap = 0;
		Set<Integer> relSet = new HashSet<Integer>();
		for (int i = 0; i < resultList.size(); ++i)
		{
			int id = resultList.get(i);
			relSet.add(truthList.get(i));
			boolean rel = relSet.contains(id);
			double precision = 0;
			for (int j = 0; j <= i; ++j)
			{
				if (relSet.contains(resultList.get(j)))
					precision += 1;
			}
			precision /= (i + 1);
			if (rel)
				ap += precision;
		}
		
		return ap / truthList.size();
	}
	
	public static double getNDCG (List<Integer> resultList, List<Integer> truthList)
	{
		double maxRel = truthList.size();
		
		SVector relVec = new SVector();
		for (int i = 0; i < truthList.size(); ++i)
			relVec.set(truthList.get(i), maxRel - i);

		double dcg = 0, idcg = 0;
		for (int i = 0; i < resultList.size(); ++i)
		{
			int id = resultList.get(i);
			double rel = relVec.get(id);
			
			double irel = maxRel - i;

			double log2i = Math.log(i + 1) / Math.log(2);
			if (i > 0)
			{
				dcg += rel / log2i;
				idcg += irel / log2i;
			}
			else
			{
				dcg += rel;
				idcg += irel;
			}
		}
		
		return dcg / idcg;
	}
	
	public static int countMismatch(List<Integer> list1, List<Integer> list2)
	{
		List<Integer> shorter, longer;
		if (list1.size() < list2.size())
		{
			shorter = list1;
			longer = list2;
		}
		else
		{
			shorter = list2;
			longer = list1;
		}
		
		int misMatch = longer.size() - shorter.size();
		assert(misMatch >= 0);
		for (int i = 0; i < shorter.size(); ++i)
		{
			int num1 = shorter.get(i);
			int num2 = longer.get(i);
			if (num1 != num2)
				misMatch++;
		}
		
		return misMatch;
	}
	
	public static double jaccardSimilarity (List<Integer> list1, List<Integer> list2)
	{
		Set<Integer> inter = new HashSet<Integer>();
		Set<Integer> union = new HashSet<Integer>();
		
		inter.addAll(list1);
		inter.retainAll(list2);
		
		if (inter.size() == 0)
			return 0;

		union.addAll(list1);
		union.addAll(list2);
		
		return ((double) inter.size()) / union.size();
	}
	
	public static int locToCellRowCol (double loc, int N, double interval)
	{
		int cell = (int) (loc / interval);
		cell = cell == N ? cell - 1 : cell;
		
		return cell;
	}
	
	public static int locToCellID (Location loc, int N, double interval)
	{
		int row = locToCellRowCol(loc.getLocx(), N, interval);
		int col = locToCellRowCol(loc.getLocy(), N, interval);
		
		return row * N + col;
	}
	
	public static Location cellIdToLoc (int cellId, int N, double interval)
	{
		int row = cellId / N;
		int col = cellId % N;
		
		double locx = interval * (0.5 + row);
		double locy = interval * (0.5 + col);
		
		return new Location(locx, locy);
	}
	
	public static void printError(String error)
	{
		System.err.println("**ERROR**: " + error);
	}
	
	
	public static void mkdir(String filePath)
	{
		File file = new File(filePath);
		if (file.isDirectory() == false)
			file.mkdirs();
	}
	
	public static double getNormalization (double value, double min, double max)
	{
		return (value - min) / (max - min);
	}
	
	public static String queryStemming (String input)
	{
		PorterStemmer stemmer = new PorterStemmer();
		String [] split = input.split(" ");
		String output = "";
		for (String word : split)
		{
			if (output.equals("") == false)
				output += " ";
			stemmer.setCurrent(word);
			stemmer.stem();
			output += stemmer.getCurrent();
		}
		System.out.println("Output = " + output);
		return output;
	}
	
	private static int checkIsNoun(String word, IDictionary dict, IStemmer stemmer)
	{
		List<String> stems = stemmer.findStems(word, POS.NOUN);
		if (stems.isEmpty())
			return 1;	//non stem
		
		word = stems.get(0);
		IIndexWord idxWord = dict.getIndexWord(word, POS.NOUN);
		if (idxWord == null)
			return 2;	//no word
		
		return 0;	//is noun
	}
	
	private static int checkIsAdj(String word, IDictionary dict, IStemmer stemmer)
	{
		List<String> stems = stemmer.findStems(word, POS.ADJECTIVE);
		if (stems.isEmpty())
			return 1;	//non stem
		
		word = stems.get(0);
		IIndexWord idxWord = dict.getIndexWord(word, POS.ADJECTIVE);
		if (idxWord == null)
			return 2;	//no word
		
		return 0;	//is noun
	}
	
	public static boolean filter(String phrase, IDictionary dict, IStemmer stemmer)
	{	
		if (phrase.length() <= 3)
			return true;
		
		String [] words = phrase.split(" ");
		String word = words[words.length - 1];	//last word
		
		for (String stopword : stopWords)
		{
			if (word.equals(stopword))
				return true;
		}
		
		int isNoun = checkIsNoun(word, dict, stemmer);
		if (isNoun == 0)
			return false;
		
		int isAdj = checkIsAdj(word, dict, stemmer);
		if (isAdj == 0)
			return false;
		
		return true;
	}
	
	public static boolean filter(String phrase)
	{	
		String [] words = phrase.split(" ");
		String word = words[words.length - 1];
		for (String stopword : stopWords)
		{
			if (word.equals(stopword))
				return true;
		}
		
		return false;
	}
	
	public static final List<String> stopWords = Arrays.asList(
  	      "a", "an", "and", "are", "as", "at", "be", "but", "by",
  	      "for", "if", "in", "into", "is", "it",
  	      "no", "not", "of", "on", "or", "such",
  	      "that", "the", "their", "then", "there", "these",
  	      "they", "this", "to", "was", "will", "with",
  	      "ar", "wa", "thei", "thi");
	//a an and ar a at be but by for if in into i it no not of on or such that the their then there these thei thi to wa will with
	
    public static final Map<String, Location> locMap;
    static
    {
    	locMap = new HashMap<String, Location>();
    	locMap.put("HOU", new Location(29.817178, -95.4012915));	//Houston
    	locMap.put("NY", new Location(40.7056307, -73.9780035));	//New York
    	locMap.put("LA", new Location(34.0204989, -118.4117325));	//Los Angeles
    	locMap.put("ZERO", new Location(0, 0));
    }
    
	public static int getMaxValue (int [] array)
	{
		int maxValue = array[0];
		for (int i = 1; i < array.length; ++i)
			maxValue = array[i] > maxValue ? array[i] : maxValue;
		return maxValue;
	}
	
	public static <T extends Comparable<T>> T getMaxValue (T [] array)
	{
		T maxValue = array[0];
		for (int i = 1; i < array.length; ++i)
			maxValue = array[i].compareTo(maxValue) > 0 ? array[i] : maxValue;
		return maxValue;
	}
	
	public static <T extends Comparable<T>> T getMinValue(T [] array)
	{
		T minValue = array[0];
		for (int i = 1; i < array.length; ++i)
			minValue = array[i].compareTo(minValue) < 0 ? array[i] : minValue;
		return minValue;
	}
	
	public static void normalizeArrayTo01(double [] array)
	{
		double maxValue = array[0];
		double minValue = array[0];
		for (int i = 1; i < array.length; ++i)
		{
			maxValue = array[i] > maxValue ? array[i] : maxValue;
			minValue = array[i] < minValue ? array[i] : minValue;
		}
		
		if (maxValue > minValue)
		{
			for (int i = 0; i < array.length; ++i)
				array[i] = (array[i] - minValue) / (maxValue - minValue);
		}
		else
		{
			// If everything is equal, set all to 1
			// this is to consider the case where gamma = 0
			for (int i = 0; i < array.length; ++i)
				array[i] = 1;
		}
	}
	
	
	public static void normalizeArrayToUnit(double [] array)
	{
		double sum = 0;
		for (int i = 0; i < array.length; ++i)
			sum += array[i];
		
		for (int i = 0; i < array.length; ++i)
			array[i] /= sum;
	}
	
	public static double distance (double ax, double ay, double bx, double by)
	{
		return Math.sqrt((ax - bx) * (ax - bx) + (ay - by) * (ay - by));
	}
	
	public static double distance (Location locA, Location locB)
	{
		return distance(locA.getLocx(), locA.getLocy(), locB.getLocx(), locB.getLocy());
	}
	
	public static void printProgress(int count, int total, double cost)
	{
		System.out.flush();
		double ratio = (double) count/total;
		System.out.printf("\r%d %.3f%% (Elapsed: %d. Remaining: %d)", count, ratio * 100, (int)cost, (int) (cost / ratio * (1 - ratio)));
//		final int width = 50;
//		System.out.print(" [");
//		int i = 0;
//		for (i = 0; i <= (int)(percentage * width); ++i)
//			System.out.print(".");
//		for (; i < width; ++i)
//			System.out.print(" ");
//		System.out.print("]");
	}
	
	public static boolean isNumeric(String str)
	{
		return str.matches("\\d+(\\.\\d+)?");  //match a number with (optionally) a dot.
	}
}
